"""---
title: "Objects - Ontology Object Instance"
slug: "sdk-ref-objects-ont-object-instance"
hidden: false
metadata:
  title: "Objects - Ontology Object Instance"
  description: "Encord SDK Objects - Ontology Object Instances."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from encord.common.range_manager import RangeManager
from encord.common.time_parser import parse_datetime
from encord.constants.enums import DataType
from encord.exceptions import LabelRowError
from encord.objects import ChecklistAttribute, RadioAttribute, Shape, TextAttribute
from encord.objects.answers import (
    Answer,
    _get_static_answer_map,
    get_default_answer_from_attribute,
)
from encord.objects.attributes import Attribute, _get_attribute_by_hash
from encord.objects.constants import DEFAULT_CONFIDENCE, DEFAULT_MANUAL_ANNOTATION
from encord.objects.coordinates import (
    ACCEPTABLE_COORDINATES_FOR_ONTOLOGY_ITEMS,
    NON_GEOMETRIC_COORDINATES,
    AudioCoordinates,
    Coordinates,
    HtmlCoordinates,
    TextCoordinates,
)
from encord.objects.frames import (
    Frames,
    Range,
    Ranges,
    frames_class_to_frames_list,
    frames_to_ranges,
    ranges_list_to_ranges,
)
from encord.objects.html_node import HtmlRange, HtmlRanges
from encord.objects.internal_helpers import (
    _infer_attribute_from_answer,
    _search_child_attributes,
)
from encord.objects.ontology_object import Object
from encord.objects.options import Option
from encord.objects.utils import check_email, short_uuid_str

if TYPE_CHECKING:
    from encord.objects.ontology_labels_impl import LabelRowV2


class ObjectInstance:
    """An object instance is an object that has coordinates and can be placed on one or multiple frames in a label row."""

    def __init__(self, ontology_object: Object, *, object_hash: Optional[str] = None):
        self._ontology_object = ontology_object
        self._object_hash = object_hash or short_uuid_str()
        self._parent: Optional[LabelRowV2] = None

        self._static_answer_map: Dict[str, Answer] = _get_static_answer_map(self._ontology_object.attributes)
        # feature_node_hash of attribute to the answer.

        self._dynamic_answer_manager = DynamicAnswerManager(self)

        # Only used for non-frame entities
        self._non_geometric = ontology_object.shape in (Shape.AUDIO, Shape.TEXT)

        self._frames_to_instance_data: Dict[int, ObjectInstance.FrameData] = {}

    def is_assigned_to_label_row(self) -> Optional[LabelRowV2]:
        """Checks if the object instance is assigned to a label row.

        Returns:
            The LabelRowV2 instance if assigned, otherwise None.
        """
        return self._parent

    @property
    def object_hash(self) -> str:
        """A unique identifier for the object instance.

        Returns:
            The unique object hash.
        """
        return self._object_hash

    @property
    def ontology_item(self) -> Object:
        """The ontology object associated with this instance.

        Returns:
            The ontology object.
        """
        return self._ontology_object

    @property
    def feature_hash(self) -> str:
        """Feature node hash from the project ontology.

        Returns:
            The feature node hash.
        """
        return self._ontology_object.feature_node_hash

    @property
    def object_name(self) -> str:
        """Object name from the project ontology.

        Returns:
            The object name.
        """
        return self._ontology_object.name

    @property
    def _last_frame(self) -> Union[int, float]:
        if self._parent is None or self._parent.data_type is DataType.DICOM:
            return float("inf")
        else:
            return self._parent.number_of_frames

    @property
    def range_list(self) -> Ranges | None:
        if self._non_geometric:
            non_geometric_annotation = self._get_non_geometric_annotation()
            if non_geometric_annotation is None:
                return None

            coordinates = non_geometric_annotation.coordinates

            if isinstance(coordinates, (AudioCoordinates, TextCoordinates)):
                return coordinates.range
            else:
                return None

        else:
            raise LabelRowError(
                "No ranges available for this object instance."
                "Please ensure the object instance was created with "
                "the range_only property to True (via a range aware shape)"
                "You can do ObjectInstance(audio_ontology_object) to achieve this."
            )

    @property
    def range_html(self) -> Optional[HtmlRanges]:
        if not self._non_geometric:
            return None

        non_geometric_annotation = self._get_non_geometric_annotation()
        if non_geometric_annotation is None:
            return None

        coordinates = non_geometric_annotation.coordinates

        if isinstance(coordinates, HtmlCoordinates):
            return coordinates.range
        else:
            return None

    def is_range_only(self) -> bool:
        return self._non_geometric

    def get_answer(
        self,
        attribute: Attribute,
        filter_answer: Union[str, Option, Iterable[Option], None] = None,
        filter_frame: Optional[int] = None,
        is_dynamic: Optional[bool] = None,
    ) -> Union[str, Option, Iterable[Option], AnswersForFrames, None]:
        """Get the answer set for a given ontology Attribute. Returns `None` if the attribute is not yet answered.

        For the ChecklistAttribute, it returns None if and only if
        the attribute is nested and the parent is unselected. Otherwise, if not yet answered it will return an empty
        list.

        Args:
            attribute: The ontology attribute to get the answer for.
            filter_answer: A filter for a specific answer value. Only applies to dynamic attributes.
            filter_frame: A filter for a specific frame. Only applies to dynamic attributes.
            is_dynamic: Optionally specify whether a dynamic answer is expected or not. This will throw if it is
                set incorrectly according to the attribute. Set this to narrow down the return type.

        Returns:
            If the attribute is static, then the answer value is returned, assuming an answer value has already been
            set. If the attribute is dynamic, the AnswersForFrames object is returned.
        """
        if attribute is None:
            attribute = self._ontology_object.attributes[0]
        elif not self._is_attribute_valid_child_of_object_instance(attribute):
            raise LabelRowError("The attribute is not a valid child of the classification.")
        elif not attribute.dynamic and not self._is_selectable_child_attribute(attribute):
            return None

        if is_dynamic is not None and is_dynamic is not attribute.dynamic:
            raise LabelRowError(
                f"The attribute is {'dynamic' if attribute.dynamic else 'static'}, but is_dynamic is set to "
                f"{is_dynamic}."
            )

        if attribute.dynamic:
            return self._dynamic_answer_manager.get_answer(attribute, filter_answer, filter_frame)

        static_answer = self._static_answer_map[attribute.feature_node_hash]

        if not static_answer.is_answered():
            if isinstance(attribute, ChecklistAttribute):
                return []
            return None

        return static_answer.get()

    def set_answer(
        self,
        answer: Union[str, Option, Sequence[Option]],
        attribute: Optional[Attribute] = None,
        frames: Optional[Frames] = None,
        overwrite: bool = False,
        manual_annotation: bool = DEFAULT_MANUAL_ANNOTATION,
    ) -> None:
        """Set the answer for a given ontology Attribute. This is the equivalent of e.g. selecting a checkbox in the
        UI after drawing the ObjectInstance. There is only one answer per ObjectInstance per Attribute, unless
        the attribute is dynamic (check the args list for more instructions on how to set dynamic answers).

        Args:
            answer: The answer to set.
            attribute: The ontology attribute to set the answer for. If not set, this will be attempted to be
                inferred.  For answers to :class:`encord.objects.common.RadioAttribute` or
                :class:`encord.objects.common.ChecklistAttribute`, this can be inferred automatically. For
                :class:`encord.objects.common.TextAttribute`, this will only be inferred there is only one possible
                TextAttribute to set for the entire object instance. Otherwise, a
                :class:`encord.exceptionsLabelRowError` will be thrown.
            frames: Only relevant for dynamic attributes. The frames to set the answer for. If `None`, the
                answer is set for all frames that this object currently has set coordinates for (also overwriting
                current answers). This will not automatically propagate the answer to new frames that are added in the
                future.
                If this is anything but `None` for non-dynamic attributes, this will
                throw a ValueError.
            overwrite: If `True`, the answer will be overwritten if it already exists. If `False`, this will throw
                a LabelRowError if the answer already exists. This argument is ignored for dynamic attributes.
            manual_annotation: If `True`, the answer will be marked as manually annotated. This arg defaults to
                DEFAULT_MANUAL_ANNOTATION.
        """
        if attribute is None:
            attribute = _infer_attribute_from_answer(self._ontology_object.attributes, answer)
        if not self._is_attribute_valid_child_of_object_instance(attribute):
            raise LabelRowError("The attribute is not a valid child of the object.")
        elif not attribute.dynamic and not self._is_selectable_child_attribute(attribute):
            raise LabelRowError(
                "Setting a nested attribute is only possible if all parent attributes have been selected."
            )
        elif frames is not None and attribute.dynamic is False:
            raise LabelRowError("Setting frames is only possible for dynamic attributes.")

        if attribute.dynamic:
            self._dynamic_answer_manager.set_answer(answer, attribute, frames)
            return

        static_answer = self._static_answer_map[attribute.feature_node_hash]
        if static_answer.is_answered() and overwrite is False:
            raise LabelRowError(
                "The answer to this attribute was already set. Set `overwrite` to `True` if you want to"
                "overwrite an existing answer to an attribute."
            )
        static_answer.set(answer, manual_annotation=manual_annotation)

    def set_answer_from_list(self, answers_list: List[Dict[str, Any]]) -> None:
        """This is a low level helper function and should usually not be used directly.

        Sets the answer for the classification from a dictionary.

        Args:
            answers_list: The list of dictionaries to set the answer from.
        """
        grouped_answers = defaultdict(list)

        for answer_dict in answers_list:
            attribute = _get_attribute_by_hash(answer_dict["featureHash"], self._ontology_object.attributes)
            if attribute is None:
                raise LabelRowError(
                    "One of the attributes does not exist in the ontology. Cannot create a valid LabelRow."
                )
            if not self._is_attribute_valid_child_of_object_instance(attribute):
                raise LabelRowError(
                    "One of the attributes set for a classification is not a valid child of the classification. "
                    "Cannot create a valid LabelRow."
                )

            grouped_answers[attribute.feature_node_hash].append(answer_dict)

        #
        # UI structures answers for checkboxes differently from SDK.
        # It has separate answer dict with the same feature hash rather than just one dict with multiple answers,
        # as SDK expects.
        # So until we aligned the models, we need to introduce additional adaptation layer, that groups separate
        # dictionaries with one answer into one with multiple answer.
        # This is a hotfix rather than a proper solution.
        # TODO: agree on a one correct way to represent the checklist response, and change the SDK accordingly.
        #

        for feature_hash, answers_list in grouped_answers.items():
            attribute = _get_attribute_by_hash(feature_hash, self._ontology_object.attributes)
            assert attribute  # we already checked that attribute is not null above. So just silencing this for now
            self._set_answer_from_grouped_list(attribute, answers_list)

    @staticmethod
    def _merge_answers_to_non_overlapping_ranges(ranges: List[Tuple[Range, Set[str]]]) -> List[Tuple[Range, Set[str]]]:
        ranges.sort(key=lambda x: x[0].start)

        edges: List[Tuple[int, bool, Set[str]]] = []
        for r, option_ids in ranges:
            edges.extend(((r.start, True, set(option_ids)), (r.end, False, set(option_ids))))
        edges.sort(key=lambda x: x[0])

        result_ranges: Dict[Tuple[int, int], Tuple[Range, Set[str]]] = {}

        prev = 0
        prev_state: Set[str] = set()
        prev_close = False

        for frame_num, is_start, options in edges:
            if is_start:
                new_state = prev_state.union(options)
                start = prev
                end = frame_num - int((len(prev_state) > 0))
                prev_close = False
            else:
                start = prev + int(prev_close)
                end = frame_num
                prev_close = True
                new_state = prev_state.difference(options)

            if len(prev_state) > 0:
                if (start, end) in result_ranges:
                    result_ranges[(start, end)][1].update(prev_state.copy())
                else:
                    result_ranges[(start, end)] = (
                        Range(start, end),
                        prev_state.copy(),
                    )

            prev_state = new_state
            prev = frame_num

        return list(result_ranges.values())

    def _set_answer_from_grouped_list(self, attribute: Attribute, answers_list: List[Dict[str, Any]]) -> None:
        if isinstance(attribute, ChecklistAttribute):
            if not attribute.dynamic:
                options = []
                for answer_dict in answers_list:
                    for answer in answer_dict["answers"]:
                        feature_hash = answer["featureHash"]
                        option = attribute.get_child_by_hash(feature_hash, type_=Option)
                        options.append(option)

                self._set_answer_unsafe(options, attribute, None)
            else:
                all_feature_hashes: Set[str] = set()
                ranges = []
                for answer_dict in answers_list:
                    feature_hashes: Set[str] = {answer["featureHash"] for answer in answer_dict["answers"]}
                    all_feature_hashes.update(feature_hashes)
                    for frame_range in ranges_list_to_ranges(answer_dict["range"]):
                        ranges.append((frame_range, feature_hashes))

                options_cache = {
                    feature_hash: attribute.get_child_by_hash(feature_hash, type_=Option)
                    for feature_hash in all_feature_hashes
                }

                for frame_range, feature_hashes in self._merge_answers_to_non_overlapping_ranges(ranges):
                    options = [options_cache[feature_hash] for feature_hash in feature_hashes]
                    self._set_answer_unsafe(options, attribute, [frame_range])
        else:
            for answer in answers_list:
                self._set_answer_from_dict(answer, attribute)

    def delete_answer(
        self,
        attribute: Attribute,
        filter_answer: Optional[Union[str, Option, Iterable[Option]]] = None,
        filter_frame: Optional[int] = None,
    ) -> None:
        """Reset the answer of an attribute as if it was never set.

        Args:
            attribute: The attribute to delete the answer for.
            filter_answer: A filter for a specific answer value. Delete only answers with the provided value.
                Only applies to dynamic attributes.
            filter_frame: A filter for a specific frame. Only applies to dynamic attributes.
        """
        if attribute.dynamic:
            self._dynamic_answer_manager.delete_answer(attribute, filter_frame, filter_answer)
            return

        static_answer = self._static_answer_map[attribute.feature_node_hash]
        static_answer.unset()

    def check_within_range(self, frame: int) -> None:
        """Check if the given frame is within the acceptable range.

        Args:
            frame: The frame number to check.

        Raises:
            LabelRowError: If the frame is out of the acceptable range.
        """
        if frame < 0 or frame >= self._last_frame:
            raise LabelRowError(
                f"The supplied frame of `{frame}` is not within the acceptable bounds of `0` to `{self._last_frame}`."
            )

    def set_for_frames(
        self,
        coordinates: Coordinates,
        frames: Frames = 0,
        *,
        overwrite: bool = False,
        created_at: Optional[datetime] = None,
        created_by: Optional[str] = None,
        last_edited_at: Optional[datetime] = None,
        last_edited_by: Optional[str] = None,
        confidence: Optional[float] = None,
        manual_annotation: Optional[bool] = None,
        reviews: Optional[List[dict]] = None,
        is_deleted: Optional[bool] = None,
    ) -> None:
        """Place the object onto the specified frame(s).

        If the object already exists on the frame and overwrite is set to `True`,
        the currently specified values will be overwritten.

        Args:
            coordinates: The coordinates of the object in the frame.
                This will throw an error if the type of the coordinates does not match the type of the attribute in the object instance.
            frames: The frames to add the object instance to. Defaults to the first frame for convenience.
            overwrite: If `True`, overwrite existing data for the given frames.
                This will not reset all the non-specified values.
                If `False` and data already exists for the given frames, raises an error.
            created_at: Optionally specify the creation time of the object instance on this frame.
                Defaults to `datetime.now()`.
            created_by: Optionally specify the creator of the object instance on this frame.
                Defaults to the current SDK user.
            last_edited_at: Optionally specify the last edit time of the object instance on this frame.
                Defaults to `datetime.now()`.
            last_edited_by: Optionally specify the last editor of the object instance on this frame.
                Defaults to the current SDK user.
            confidence: Optionally specify the confidence of the object instance on this frame. Defaults to `1.0`.
            manual_annotation: Optionally specify whether the object instance on this frame was manually annotated. Defaults to `True`.
            reviews: Should only be set by internal functions.
            is_deleted: Should only be set by internal functions.
        """
        if self._non_geometric:
            if not isinstance(coordinates, tuple(NON_GEOMETRIC_COORDINATES)):
                raise LabelRowError("Expecting non-geometric coordinate type")

            elif frames != 0:
                raise LabelRowError(
                    f"For objects with a non-geometric shape (e.g. {Shape.TEXT} and {Shape.AUDIO}), "
                    f"There is only one frame. Please ensure `set_for_frames` is called with `frames=0`."
                )

        frames_list = frames_class_to_frames_list(frames)

        for frame in frames_list:
            existing_frame_data = self._frames_to_instance_data.get(frame)

            if overwrite is False and existing_frame_data is not None:
                raise LabelRowError(
                    "Cannot overwrite existing data for a frame. Set `overwrite` to `True` to overwrite."
                )

            check_coordinate_type(coordinates, self._ontology_object, self._parent)

            if isinstance(coordinates, (TextCoordinates, AudioCoordinates)):
                for non_geometric_range in coordinates.range:
                    self.check_within_range(non_geometric_range.end)
            else:
                self.check_within_range(frame)

            if existing_frame_data is None:
                existing_frame_data = ObjectInstance.FrameData(
                    coordinates=coordinates, object_frame_instance_info=ObjectInstance.FrameInfo()
                )
                self._frames_to_instance_data[frame] = existing_frame_data

            existing_frame_data.object_frame_instance_info.update_from_optional_fields(
                created_at=created_at,
                created_by=created_by,
                last_edited_at=last_edited_at,
                last_edited_by=last_edited_by,
                confidence=confidence,
                manual_annotation=manual_annotation,
                reviews=reviews,
                is_deleted=is_deleted,
            )
            existing_frame_data.coordinates = coordinates

            if self._parent:
                self._parent.add_to_single_frame_to_hashes_map(self, frame)

    def _get_non_geometric_annotation(self) -> Optional[Annotation]:
        # Non-geometric annotations (e.g. Audio and Text) only have one frame.
        if 0 not in self._frames_to_instance_data:
            return None
        else:
            return self.get_annotation(0)

    def get_annotation(self, frame: Union[int, str] = 0) -> Annotation:
        """Get the annotation for the object instance on the specified frame.

        Args:
            frame: Either the frame number or the image hash if the data type is an image or image group.
                Defaults to the first frame.

        Returns:
            Annotation: The annotation for the specified frame.

        Raises:
            LabelRowError: If the frame is not present in the label row.
        """
        if self._non_geometric and frame != 0:
            raise LabelRowError(
                'This annotation data for this object instance is stored on only one "frame". '
                "Use `get_annotation(0)` to get the frame data of the first frame."
            )
        if isinstance(frame, str):
            # TODO: this check should be consistent for both string and integer frames,
            #       but currently it is not possible due to the parsing logic
            if not self._parent:
                raise LabelRowError("Cannot get annotation for an object instance that is not assigned to a label row.")

            frame_num = self._parent.get_frame_number(frame)
            if frame_num is None:
                raise LabelRowError(f"Image hash {frame} is not present in the label row.")
        else:
            frame_num = frame

        return self.Annotation(self, frame_num)

    def copy(self) -> ObjectInstance:
        """Create an exact copy of this ObjectInstance.

        The new copy will have a new object hash and will not be associated with any `LabelRowV2`.
        This is useful for adding the semantically same ObjectInstance to multiple `LabelRowV2`s.

        Returns:
            ObjectInstance: A new ObjectInstance that is a copy of the current instance.
        """
        ret = ObjectInstance(self._ontology_object)
        ret._frames_to_instance_data = deepcopy(self._frames_to_instance_data)
        ret._static_answer_map = deepcopy(self._static_answer_map)
        ret._dynamic_answer_manager = self._dynamic_answer_manager.copy()
        return ret

    def get_annotations(self) -> List[Annotation]:
        """Get all annotations for the object instance on all frames it has been placed on.

        Returns:
            List[Annotation]: A list of `ObjectInstance.Annotation` in order of available frames.
        """
        return [self.get_annotation(frame_num) for frame_num in sorted(self._frames_to_instance_data.keys())]

    def get_annotation_frames(self) -> set[int]:
        """Get all annotations for the object instance on all frames it has been placed on.

        Returns:
            List[Annotation]: A list of `ObjectInstance.Annotation` in order of available frames.
        """
        return {self.get_annotation(frame_num).frame for frame_num in sorted(self._frames_to_instance_data.keys())}

    def remove_from_frames(self, frames: Frames) -> None:
        """Remove the object instance from the specified frames.

        Args:
            frames: The frames from which to remove the object instance.
        """
        if self._non_geometric and frames != 0:
            raise LabelRowError(
                f"For objects with a non-geometric shape (e.g. {Shape.TEXT} and {Shape.AUDIO}), "
                f"There is only one frame. Please ensure `remove_from_frames` is called with `frames=0`."
            )

        frames_list = frames_class_to_frames_list(frames)
        for frame in frames_list:
            self._frames_to_instance_data.pop(frame)

        if self._parent:
            self._parent._remove_from_frame_to_hashes_map(frames_list, self.object_hash)

    def is_valid(self) -> None:
        """Check if the ObjectInstance is valid.

        Raises:
            LabelRowError: If the ObjectInstance is not on any frames.
        """
        if len(self._frames_to_instance_data) == 0:
            raise LabelRowError("ObjectInstance is not on any frames. Please add it to at least one frame.")

        self.are_dynamic_answers_valid()

    def are_dynamic_answers_valid(self) -> None:
        """Validate if there are any dynamic answers on frames that have no coordinates.

        Raises:
            LabelRowError: If there are dynamic answers on frames without coordinates.
        """
        dynamic_frames = set(self._dynamic_answer_manager.frames())
        local_frames = self.get_annotation_frames()

        if not len(dynamic_frames - local_frames) == 0:
            raise LabelRowError(
                "There are some dynamic answers on frames that have no coordinates. "
                "Please ensure that all the dynamic answers are only on frames where coordinates "
                "have been set previously."
            )

    class Annotation:
        """Represents an annotation for a specific frame of an ObjectInstance.

        Allows setting or getting data for the ObjectInstance on the given frame number.
        """

        def __init__(self, object_instance: ObjectInstance, frame: int):
            self._object_instance = object_instance
            self._frame = frame

        @property
        def frame(self) -> int:
            return self._frame

        @property
        def coordinates(self) -> Coordinates:
            self._check_if_frame_view_is_valid()
            return self._get_object_frame_instance_data().coordinates

        @coordinates.setter
        def coordinates(self, coordinates: Coordinates) -> None:
            self._check_if_frame_view_is_valid()
            self._object_instance.set_for_frames(coordinates, self._frame, overwrite=True)

        @property
        def created_at(self) -> datetime:
            self._check_if_frame_view_is_valid()
            return self._get_object_frame_instance_data().object_frame_instance_info.created_at

        @created_at.setter
        def created_at(self, created_at: datetime) -> None:
            self._check_if_frame_view_is_valid()
            self._get_object_frame_instance_data().object_frame_instance_info.created_at = created_at

        @property
        def created_by(self) -> Optional[str]:
            self._check_if_frame_view_is_valid()
            return self._get_object_frame_instance_data().object_frame_instance_info.created_by

        @created_by.setter
        def created_by(self, created_by: Optional[str]) -> None:
            """Set the created_by field with a user email or None if it should default to the current user of the SDK."""
            self._check_if_frame_view_is_valid()
            if created_by is not None:
                check_email(created_by)
            self._get_object_frame_instance_data().object_frame_instance_info.created_by = created_by

        @property
        def last_edited_at(self) -> datetime:
            self._check_if_frame_view_is_valid()
            return self._get_object_frame_instance_data().object_frame_instance_info.last_edited_at

        @last_edited_at.setter
        def last_edited_at(self, last_edited_at: datetime) -> None:
            self._check_if_frame_view_is_valid()
            self._get_object_frame_instance_data().object_frame_instance_info.last_edited_at = last_edited_at

        @property
        def last_edited_by(self) -> Optional[str]:
            self._check_if_frame_view_is_valid()
            return self._get_object_frame_instance_data().object_frame_instance_info.last_edited_by

        @last_edited_by.setter
        def last_edited_by(self, last_edited_by: Optional[str]) -> None:
            """Set the last_edited_by field with a user email or None if it should default to the current user of the SDK."""
            self._check_if_frame_view_is_valid()
            if last_edited_by is not None:
                check_email(last_edited_by)
            self._get_object_frame_instance_data().object_frame_instance_info.last_edited_by = last_edited_by

        @property
        def confidence(self) -> float:
            self._check_if_frame_view_is_valid()
            return self._get_object_frame_instance_data().object_frame_instance_info.confidence

        @confidence.setter
        def confidence(self, confidence: float) -> None:
            self._check_if_frame_view_is_valid()
            self._get_object_frame_instance_data().object_frame_instance_info.confidence = confidence

        @property
        def manual_annotation(self) -> bool:
            self._check_if_frame_view_is_valid()
            return self._get_object_frame_instance_data().object_frame_instance_info.manual_annotation

        @manual_annotation.setter
        def manual_annotation(self, manual_annotation: bool) -> None:
            self._check_if_frame_view_is_valid()
            self._get_object_frame_instance_data().object_frame_instance_info.manual_annotation = manual_annotation

        @property
        def reviews(self) -> Optional[List[Dict[str, Any]]]:
            """Get the reviews for this object on this frame.

            Returns:
                Optional[List[Dict[str, Any]]]: A list of review dictionaries, if any.
            """
            self._check_if_frame_view_is_valid()
            return self._get_object_frame_instance_data().object_frame_instance_info.reviews

        @property
        def is_deleted(self) -> Optional[bool]:
            """Check if the object instance is marked as deleted on this frame.

            Returns:
                Optional[bool]: `True` if deleted, `False` otherwise, or `None` if not set.
            """
            self._check_if_frame_view_is_valid()
            return self._get_object_frame_instance_data().object_frame_instance_info.is_deleted

        def _get_object_frame_instance_data(self) -> ObjectInstance.FrameData:
            return self._object_instance._frames_to_instance_data[self._frame]

        def _check_if_frame_view_is_valid(self) -> None:
            if self._frame not in self._object_instance._frames_to_instance_data:
                raise LabelRowError(
                    "Trying to use an ObjectInstance.Annotation for an ObjectInstance that is not on the frame"
                )

    @dataclass
    class FrameInfo:
        """Contains metadata information about a frame."""

        created_at: datetime = field(default_factory=datetime.now)
        created_by: Optional[str] = None
        """None defaults to the user of the SDK once uploaded to the server."""
        last_edited_at: datetime = field(default_factory=datetime.now)
        last_edited_by: Optional[str] = None
        """None defaults to the user of the SDK once uploaded to the server."""
        confidence: float = DEFAULT_CONFIDENCE
        manual_annotation: bool = DEFAULT_MANUAL_ANNOTATION
        reviews: Optional[List[dict]] = None
        is_deleted: Optional[bool] = None

        @staticmethod
        def from_dict(d: dict) -> ObjectInstance.FrameInfo:
            """Create a FrameInfo instance from a dictionary.

            Args:
                d: A dictionary containing frame information.

            Returns:
                ObjectInstance.FrameInfo: An instance of FrameInfo.
            """
            if "lastEditedAt" in d and d["lastEditedAt"] is not None:
                last_edited_at = parse_datetime(d["lastEditedAt"])
            else:
                last_edited_at = datetime.now()

            return ObjectInstance.FrameInfo(
                created_at=parse_datetime(d["createdAt"]),
                created_by=d["createdBy"],
                last_edited_at=last_edited_at,
                last_edited_by=d.get("lastEditedBy"),
                confidence=d["confidence"],
                manual_annotation=d.get("manualAnnotation", True),
                reviews=d.get("reviews"),
                is_deleted=d.get("isDeleted"),
            )

        def update_from_optional_fields(
            self,
            created_at: Optional[datetime] = None,
            created_by: Optional[str] = None,
            last_edited_at: Optional[datetime] = None,
            last_edited_by: Optional[str] = None,
            confidence: Optional[float] = None,
            manual_annotation: Optional[bool] = None,
            reviews: Optional[List[Dict[str, Any]]] = None,
            is_deleted: Optional[bool] = None,
        ) -> None:
            """Update the FrameInfo fields with the specified values.

            Args:
                created_at: Optional creation time.
                created_by: Optional creator identifier.
                last_edited_at: Optional last edited time.
                last_edited_by: Optional last editor identifier.
                confidence: Optional confidence value.
                manual_annotation: Optional manual annotation flag.
                reviews: Optional list of reviews.
                is_deleted: Optional deleted flag.
            """
            self.created_at = created_at or self.created_at
            if created_by is not None:
                self.created_by = created_by
            self.last_edited_at = last_edited_at or self.last_edited_at
            if last_edited_by is not None:
                self.last_edited_by = last_edited_by
            if confidence is not None:
                self.confidence = confidence
            if manual_annotation is not None:
                self.manual_annotation = manual_annotation
            if reviews is not None:
                self.reviews = reviews
            if is_deleted is not None:
                self.is_deleted = is_deleted

    @dataclass
    class FrameData:
        """Data class for storing frame-specific data.

        Attributes:
            coordinates (Coordinates): The coordinates associated with the frame.
            object_frame_instance_info (ObjectInstance.FrameInfo): The frame's metadata information.
        """

        coordinates: Coordinates
        object_frame_instance_info: ObjectInstance.FrameInfo
        # Probably the above can be flattened out into this class.

    def _set_answer_unsafe(
        self,
        answer: Union[str, Option, Iterable[Option]],
        attribute: Attribute,
        ranges: Optional[Ranges],
    ) -> None:
        if attribute.dynamic:
            self._dynamic_answer_manager.set_answer(answer, attribute, frames=ranges)
        else:
            static_answer = self._static_answer_map[attribute.feature_node_hash]
            static_answer.set(answer)

    def _set_answer_from_dict(self, answer_dict: Dict[str, Any], attribute: Attribute) -> None:
        if attribute.dynamic:
            ranges = ranges_list_to_ranges(answer_dict["range"])
        else:
            ranges = None

        if isinstance(attribute, TextAttribute):
            self._set_answer_unsafe(answer_dict["answers"], attribute, ranges)
        elif isinstance(attribute, RadioAttribute):
            if len(answer_dict["answers"]) == 1:
                feature_hash = answer_dict["answers"][0]["featureHash"]
                option = attribute.get_child_by_hash(feature_hash, type_=Option)
                self._set_answer_unsafe(option, attribute, ranges)
        elif isinstance(attribute, ChecklistAttribute):
            options = []
            for answer in answer_dict["answers"]:
                feature_hash = answer["featureHash"]
                option = attribute.get_child_by_hash(feature_hash, type_=Option)
                options.append(option)
            self._set_answer_unsafe(options, attribute, ranges)
        else:
            raise NotImplementedError(f"The attribute type {type(attribute)} is not supported.")

    def _is_attribute_valid_child_of_object_instance(self, attribute: Attribute) -> bool:
        is_static_child = attribute.feature_node_hash in self._static_answer_map
        is_dynamic_child = self._dynamic_answer_manager.is_valid_dynamic_attribute(attribute)
        return is_dynamic_child or is_static_child

    def _is_selectable_child_attribute(self, attribute: Attribute) -> bool:
        ontology_object = self._ontology_object
        for search_attribute in ontology_object.attributes:
            if search_attribute.dynamic:
                continue

            if _search_child_attributes(attribute, search_attribute, self._static_answer_map):
                return True
        return False

    def _get_all_static_answers(self) -> List[Answer]:
        return list(self._static_answer_map.values())

    def _get_all_dynamic_answers(self) -> List[Tuple[Answer, Ranges]]:
        return self._dynamic_answer_manager.get_all_answers()

    def __repr__(self) -> str:
        return (
            f"ObjectInstance(object_hash={self._object_hash}, object_name={self._ontology_object.name}, "
            f"feature_hash={self._ontology_object.feature_node_hash})"
        )

    def __hash__(self) -> int:
        return hash(id(self))

    def __lt__(self, other: ObjectInstance) -> bool:
        return self._object_hash < other._object_hash


def check_coordinate_type(coordinates: Coordinates, ontology_object: Object, parent: Optional[LabelRowV2]) -> None:
    """Check if the coordinate type matches the expected type for the ontology object.

    Args:
        coordinates (Coordinates): The coordinates to check.
        ontology_object (Object): The ontology object to check against.
        parent (LabelRowV2): The parent label row (if any) of the ontology object.

    Raises:
        LabelRowError: If the coordinate type does not match the expected type.
    """
    expected_coordinate_types = ACCEPTABLE_COORDINATES_FOR_ONTOLOGY_ITEMS[ontology_object.shape]
    if all(
        not isinstance(coordinates, expected_coordinate_type) for expected_coordinate_type in expected_coordinate_types
    ):
        raise LabelRowError(
            f"Expected coordinates of one of the following types: `{expected_coordinate_types}`, but got type `{type(coordinates)}`."
        )

    # An ontology object with `Text` shape can have both coordinates `HtmlCoordinates` and `TextCoordinates`
    # Therefore, we need to further check the file type, to ensure that `HtmlCoordinates` are only used for
    # HTML files, and `TextCoordinates` are only used for plain text files.
    if isinstance(coordinates, TextCoordinates):
        if parent is not None and parent == "text/html":
            raise LabelRowError(f"Expected coordinates of type {HtmlCoordinates}`, but got type `{type(coordinates)}`.")
    elif isinstance(coordinates, HtmlCoordinates):
        if parent is not None and parent.file_type != "text/html":
            raise LabelRowError(
                "For non-html labels, ensure the `range` property " "is set when instantiating the TextCoordinates."
            )


class DynamicAnswerManager:
    """Manages dynamic answers for different frames of an ObjectInstance.

    This class is an internal helper class and should not be interacted with directly by the user.
    """

    def __init__(self, object_instance: ObjectInstance):
        self._object_instance = object_instance
        self._frames_to_answers: Dict[int, Set[Answer]] = defaultdict(set)
        self._answers_to_frames: Dict[Answer, Set[int]] = defaultdict(set)
        self._dynamic_uninitialised_answer_options: Set[Answer] = self._get_dynamic_answers()
        # ^ these are like the static answers. Everything that is possibly an answer. However,
        # don't forget also nested-ness. In this case nested-ness should be ignored.
        # ^ I might not need this object but only need the _get_dynamic_answers object.

    def is_valid_dynamic_attribute(self, attribute: Attribute) -> bool:
        """Check if the attribute is a valid dynamic attribute.

        Args:
            attribute (Attribute): The attribute to check.

        Returns:
            bool: True if the attribute is valid, False otherwise.
        """
        return any(
            answer.ontology_attribute.feature_node_hash == attribute.feature_node_hash
            for answer in self._dynamic_uninitialised_answer_options
        )

    def delete_answer(
        self,
        attribute: Attribute,
        frames: Optional[Frames] = None,
        filter_answer: Union[str, Option, Iterable[Option], None] = None,
    ) -> None:
        """Delete the answer for a given attribute and frames.

        Args:
            attribute (Attribute): The attribute to delete the answer for.
            frames (Optional[Frames]): The frames to delete the answer for.
            filter_answer (Union[str, Option, Iterable[Option], None]): The specific answer to delete.
        """
        if frames is None:
            frames = [Range(i, i) for i in self._frames_to_answers.keys()]
        frame_list = frames_class_to_frames_list(frames)

        for frame in frame_list:
            to_remove_answer = None
            for answer_object in self._frames_to_answers[frame]:
                if filter_answer is not None:
                    if answer_object.is_answered() and answer_object.get() != filter_answer:
                        continue

                # ideally this would not be a log(n) operation, however these will not be extremely large.
                if answer_object.ontology_attribute == attribute:
                    to_remove_answer = answer_object
                    break

            if to_remove_answer is not None:
                self._frames_to_answers[frame].remove(to_remove_answer)
                if self._frames_to_answers[frame] == set():
                    del self._frames_to_answers[frame]
                self._answers_to_frames[to_remove_answer].remove(frame)
                if self._answers_to_frames[to_remove_answer] == set():
                    del self._answers_to_frames[to_remove_answer]

    def set_answer(
        self, answer: Union[str, Option, Iterable[Option]], attribute: Attribute, frames: Optional[Frames] = None
    ) -> None:
        """Set the answer for a given attribute and frames.

        Args:
            answer (Union[str, Option, Iterable[Option]]): The answer to set.
            attribute (Attribute): The attribute to set the answer for.
            frames (Optional[Frames]): The frames to set the answer for.
        """
        if frames is None:
            for available_frame_view in self._object_instance.get_annotations():
                self._set_answer(answer, attribute, available_frame_view.frame)
            return
        self._set_answer(answer, attribute, frames)

    def _set_answer(self, answer: Union[str, Option, Iterable[Option]], attribute: Attribute, frames: Frames) -> None:
        frame_list = frames_class_to_frames_list(frames)
        for frame in frame_list:
            self._object_instance.check_within_range(frame)

        self.delete_answer(attribute, frames)

        default_answer = get_default_answer_from_attribute(attribute)
        default_answer.set(answer)

        frame_list = frames_class_to_frames_list(frames)
        for frame in frame_list:
            self._frames_to_answers[frame].add(default_answer)
            self._answers_to_frames[default_answer].add(frame)

    def get_answer(
        self,
        attribute: Attribute,
        filter_answer: Union[str, Option, Iterable[Option], None] = None,
        filter_frames: Optional[Frames] = None,
    ) -> AnswersForFrames:
        """Get answers for a given attribute, filtered by the specified criteria.

        Args:
            attribute (Attribute): The attribute to get the answers for.
            filter_answer (Union[str, Option, Iterable[Option], None]): The specific answer to filter by.
            filter_frames (Optional[Frames]): The specific frames to filter by.

        Returns:
            AnswersForFrames: A list of answers and their associated frames.
        """
        ret = []
        filter_frames_set = None if filter_frames is None else set(frames_class_to_frames_list(filter_frames))
        for answer in self._answers_to_frames:
            if answer.ontology_attribute != attribute:
                continue
            if not answer.is_answered():
                continue
            if not (filter_answer is None or filter_answer == answer.get()):
                continue
            actual_frames = self._answers_to_frames[answer]
            if not (filter_frames_set is None or len(actual_frames & filter_frames_set) > 0):
                continue

            ranges = frames_to_ranges(self._answers_to_frames[answer])
            ret.append(AnswerForFrames(answer=answer.get(), ranges=ranges))
        return ret

    def frames(self) -> Iterable[int]:
        """Get all frames that have answers set.

        Returns:
            Iterable[int]: An iterable of frames.
        """
        return self._frames_to_answers.keys()

    def get_all_answers(self) -> List[Tuple[Answer, Ranges]]:
        """Get all answers that are set.

        Returns:
            List[Tuple[Answer, Ranges]]: A list of tuples containing the answer and its associated ranges.
        """
        return [(answer, frames_to_ranges(frames)) for answer, frames in self._answers_to_frames.items()]

    def copy(self) -> DynamicAnswerManager:
        """Create a deep copy of the DynamicAnswerManager instance.

        Returns:
            DynamicAnswerManager: A new instance of DynamicAnswerManager with copied data.
        """
        ret = DynamicAnswerManager(self._object_instance)
        ret._frames_to_answers = deepcopy(self._frames_to_answers)
        ret._answers_to_frames = deepcopy(self._answers_to_frames)
        return ret

    def _get_dynamic_answers(self) -> Set[Answer]:
        ret: Set[Answer] = set()
        for attribute in self._object_instance.ontology_item.attributes:
            if attribute.dynamic:
                answer = get_default_answer_from_attribute(attribute)
                ret.add(answer)
        return ret

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DynamicAnswerManager):
            return False
        return (
            self._frames_to_answers == other._frames_to_answers and self._answers_to_frames == other._answers_to_frames
        )

    def __hash__(self) -> int:
        return hash(id(self))


@dataclass
class AnswerForFrames:
    """Data class for storing an answer and its associated frame ranges.

    Attributes:
        answer (Union[str, Option, Iterable[Option]]): The answer set for the frames.
        ranges (Ranges): The ranges representing the frames where the answer is set.

    The ranges are essentially a run-length encoding of the frames where the unique answer is set.
    They are sorted in ascending order.
    """

    answer: Union[str, Option, Iterable[Option]]
    ranges: Ranges


AnswersForFrames = List[AnswerForFrames]
"""
A list of AnswerForFrames objects, representing answers and their associated frame ranges.
"""



from enum import Enum
from typing import List
from uuid import UUID

from encord.orm.base_dto import BaseDTO


class WorkflowAction(str, Enum):
    REOPEN = "reopen"
    COMPLETE = "complete"


class WorkflowStageType(str, Enum):
    ANNOTATION = "ANNOTATION"
    REVIEW = "REVIEW"
    USER_ROUTER = "USER_ROUTER"
    PERCENTAGE_ROUTER = "PERCENTAGE_ROUTER"
    CONSENSUS_ANNOTATION = "CONSENSUS_ANNOTATION"
    CONSENSUS_REVIEW = "CONSENSUS_REVIEW"
    DONE = "DONE"
    AGENT = "AGENT"


class LabelWorkflowGraphNode:
    pass


class LabelWorkflowGraphNodePayload(BaseDTO):
    action: str


class WorkflowNode(BaseDTO):
    uuid: UUID
    stage_type: WorkflowStageType
    title: str


class Workflow(BaseDTO):
    stages: List[WorkflowNode]



import logging
from datetime import datetime
from typing import Iterator, List, Optional, Sequence, Tuple, Union
from uuid import UUID

import encord.orm.storage as orm_storage
from encord.client import EncordClientProject
from encord.exceptions import (
    AuthorisationError,
)
from encord.filter_preset import FilterPreset
from encord.http.v2.api_client import ApiClient
from encord.objects.ontology_labels_impl import LabelRowV2
from encord.ontology import Ontology
from encord.orm.collection import Collection as OrmCollection
from encord.orm.collection import (
    CollectionBulkItemRequest,
    CollectionBulkItemResponse,
    CollectionBulkPresetRequest,
    CreateCollectionParams,
    CreateCollectionPayload,
    CreateProjectCollectionParams,
    CreateProjectCollectionPayload,
    GetCollectionItemsParams,
    GetCollectionParams,
    GetCollectionsResponse,
    GetProjectCollectionParams,
    ProjectCollectionBulkItemRequest,
    ProjectCollectionBulkItemResponse,
    ProjectCollectionType,
    ProjectDataCollectionInstance,
    ProjectDataCollectionItemRequest,
    ProjectDataCollectionItemResponse,
    ProjectLabelCollectionInstance,
    ProjectLabelCollectionItemRequest,
    ProjectLabelCollectionItemResponse,
    UpdateCollectionPayload,
)
from encord.orm.collection import ProjectCollection as OrmProjectCollection
from encord.orm.label_row import LabelRowMetadata, label_row_metadata_dto_to_label_row_metadata
from encord.storage import StorageItem, StorageItemInaccessible

log = logging.getLogger(__name__)


class Collection:
    """Represents collections in Index.
    Collections are a logical grouping of data items that can be used to
    create datasets and perform various data curation flows.
    """

    def __init__(self, client: ApiClient, orm_collection: OrmCollection):
        self._client = client
        self._collection_instance = orm_collection

    @property
    def uuid(self) -> UUID:
        """Get the collection unique identifier (UUID).

        Returns:
            UUID: The collection UUID.
        """
        return self._collection_instance.uuid

    @property
    def name(self) -> str:
        """Get the collection name.

        Returns:
            str: The collection name.
        """
        return self._collection_instance.name

    @property
    def description(self) -> Optional[str]:
        """Get the collection description.

        Returns:
            Optional[str]: The collection description, or None if not available.
        """
        return self._collection_instance.description

    @property
    def created_at(self) -> Optional[datetime]:
        """Get the collection creation timestamp.

        Returns:
            Optional[datetime]: The timestamp when the collection was created, or None if not available.
        """
        return self._collection_instance.created_at

    @property
    def last_edited_at(self) -> Optional[datetime]:
        """Get the collection last edit timestamp.

        Returns:
            Optional[datetime]: The timestamp when the collection was last edited, or None if not available.
        """
        return self._collection_instance.last_edited_at

    @property
    def top_level_folder_uuid(self) -> UUID:
        """Get the uuid of the top level folder that the collection is on.

        Returns:
            UUID: the uuid of the top level folder that the collection is on
        """
        return self._collection_instance.top_level_folder_uuid

    @staticmethod
    def _get_collection(api_client: ApiClient, collection_uuid: UUID) -> "Collection":
        params = GetCollectionParams(uuids=[collection_uuid])
        orm_item = api_client.get(
            "index/collections",
            params=params,
            result_type=GetCollectionsResponse,
        )
        if len(orm_item.results) > 0:
            return Collection(api_client, orm_item.results[0])
        raise AuthorisationError("No collection found")

    @staticmethod
    def _list_collections(
        api_client: ApiClient,
        top_level_folder_uuid: Union[UUID, None],
        collection_uuids: Union[List[UUID], None],
        page_size: Optional[int] = None,
    ) -> Iterator["Collection"]:
        params = GetCollectionParams(
            topLevelFolderUuid=top_level_folder_uuid, uuids=collection_uuids, pageSize=page_size
        )
        paged_items = api_client.get_paged_iterator(
            "index/collections",
            params=params,
            result_type=OrmCollection,
        )
        for item in paged_items:
            yield Collection(api_client, item)

    @staticmethod
    def _delete_collection(api_client: ApiClient, collection_uuid: UUID) -> None:
        api_client.delete(
            f"index/collections/{collection_uuid}",
            params=None,
            result_type=None,
        )

    @staticmethod
    def _create_collection(
        api_client: ApiClient, top_level_folder_uuid: UUID, name: str, description: str = ""
    ) -> UUID:
        params = CreateCollectionParams(topLevelFolderUuid=top_level_folder_uuid)
        payload = CreateCollectionPayload(name=name, description=description)
        orm_item = api_client.post(
            "index/collections",
            params=params,
            payload=payload,
            result_type=UUID,
        )
        return orm_item

    def update_collection(self, name: Optional[str] = None, description: Optional[str] = None) -> None:
        """Update the collection's name and/or description.

        Args:
           name (Optional[str]): The new name for the collection.
           description (Optional[str]): The new description for the collection.
        """
        payload = UpdateCollectionPayload(name=name, description=description)
        self._client.patch(
            f"index/collections/{self.uuid}",
            params=None,
            payload=payload,
            result_type=None,
        )

    def list_items(
        self,
        page_size: Optional[int] = None,
    ) -> Iterator[StorageItem]:
        """List storage items in the collection.

        Args:
            page_size (Optional[int]): The number of items to fetch per page.

        Returns:
            Iterator[StorageItem]: An iterator containing storage items in the collection.
        """
        params = GetCollectionItemsParams(pageSize=page_size)
        paged_items = self._client.get_paged_iterator(
            f"index/collections/{self.uuid}/accessible-items",
            params=params,
            result_type=orm_storage.StorageItem,
        )
        for item in paged_items:
            yield StorageItem(api_client=self._client, orm_item=item)

    def list_items_include_inaccessible(
        self, page_size: Optional[int] = None
    ) -> Iterator[Union[StorageItem, StorageItemInaccessible]]:
        """List storage items in the collection, including those that are inaccessible.

        Args:
            page_size (Optional[int]): The number of items to fetch per page.

        Returns:
            Iterator[Union[StorageItem, StorageItemInaccessible]]: An iterator containing both accessible
            and inaccessible storage items in the collection.
        """
        params = GetCollectionItemsParams(pageSize=page_size)
        paged_items: Iterator[Union[orm_storage.StorageItem, orm_storage.StorageItemInaccessible]] = (
            self._client.get_paged_iterator(
                f"index/collections/{self.uuid}/all-items",
                params=params,
                result_type=Union[orm_storage.StorageItem, orm_storage.StorageItemInaccessible],  # type: ignore[arg-type]
            )
        )
        for item in paged_items:
            if isinstance(item, orm_storage.StorageItem):
                yield StorageItem(api_client=self._client, orm_item=item)
            else:
                yield StorageItemInaccessible(orm_item=item)

    def add_items(self, storage_item_uuids: Sequence[Union[UUID, str]]) -> CollectionBulkItemResponse:
        """Add storage items to the collection.

        Args:
            storage_item_uuids (Sequence[Union[UUID, str]]): The list of storage item UUIDs to be added.
            Either UUIDs or string representations of UUIDs are accepted.

        Returns:
            CollectionBulkItemResponse: The response after adding items to the collection.
        """
        uuid_list = [item if isinstance(item, UUID) else UUID(item) for item in storage_item_uuids]
        res = self._client.post(
            f"index/collections/{self.uuid}/add-items",
            params=None,
            payload=CollectionBulkItemRequest(item_uuids=uuid_list),
            result_type=CollectionBulkItemResponse,
        )
        return res

    def remove_items(self, storage_item_uuids: Sequence[Union[UUID, str]]) -> CollectionBulkItemResponse:
        """Remove storage items from the collection.

        Args:
            storage_item_uuids (Sequence[Union[UUID, str]]): The list of storage item UUIDs to be removed.
            Either UUIDs or string representations of UUIDs are accepted.

        Returns:
            CollectionBulkItemResponse: The response after removing items from the collection.
        """
        uuid_list = [item if isinstance(item, UUID) else UUID(item) for item in storage_item_uuids]
        res = self._client.post(
            f"index/collections/{self.uuid}/remove-items",
            params=None,
            payload=CollectionBulkItemRequest(item_uuids=uuid_list),
            result_type=CollectionBulkItemResponse,
        )
        return res

    def add_preset_items(self, filter_preset: Union[FilterPreset, UUID, str]) -> None:
        """Async operation to add storage items matching a filter preset to the collection.

        Args:
             filter_preset (Union[FilterPreset, UUID, str]): The filter preset or its UUID/ID used to filter items.
        """
        if isinstance(filter_preset, FilterPreset):
            preset_uuid = filter_preset.uuid
        elif isinstance(filter_preset, str):
            preset_uuid = UUID(filter_preset)
        else:
            preset_uuid = filter_preset
        self._client.post(
            f"index/collections/{self.uuid}/add-preset-items",
            params=None,
            payload=CollectionBulkPresetRequest(preset_uuid=preset_uuid),
            result_type=None,
        )
        log.info(
            f"Submitted request to add items matching filter_preset:{preset_uuid} to collection:{self.uuid}."
            f"It is an async operation and can take some time to complete."
        )

    def remove_preset_items(self, filter_preset: Union[FilterPreset, UUID, str]) -> None:
        """Async operation to remove storage items matching a filter preset from the collection.

        Args:
            filter_preset (Union[FilterPreset, UUID, str]): The filter preset or its UUID/ID used to filter items.
        """
        if isinstance(filter_preset, FilterPreset):
            preset_uuid = filter_preset.uuid
        elif isinstance(filter_preset, str):
            preset_uuid = UUID(filter_preset)
        else:
            preset_uuid = filter_preset
        self._client.post(
            f"index/collections/{self.uuid}/remove-preset-items",
            params=None,
            payload=CollectionBulkPresetRequest(preset_uuid=preset_uuid),
            result_type=None,
        )
        log.info(
            f"Submitted request to remove items matching filter_preset:{preset_uuid} from collection:{self.uuid}."
            f"It is an async operation and can take some time to complete."
        )


class ProjectCollection:
    """Represents Active collections inside a Project.
    Active Project Collections are a logical grouping of frames (images or video frames) or
    annotations (objects and classifications) that can be used to perform various data curation flows.
    """

    def __init__(
        self,
        project_uuid: UUID,
        project_client: EncordClientProject,
        ontology: Ontology,
        orm_collection: OrmProjectCollection,
    ):
        self._project_uuid = project_uuid
        self._client = project_client._api_client
        self._project_client = project_client
        self._ontology = ontology
        self._collection_instance = orm_collection

    @property
    def uuid(self) -> UUID:
        """Get the collection unique identifier (UUID).

        Returns:
            UUID: The collection UUID.
        """
        return self._collection_instance.collection_uuid

    @property
    def name(self) -> str:
        """Get the collection name.

        Returns:
            str: The collection name.
        """
        return self._collection_instance.name

    @property
    def description(self) -> Optional[str]:
        """Get the collection description.

        Returns:
            Optional[str]: The collection description, or None if not available.
        """
        return self._collection_instance.description

    @property
    def created_at(self) -> Optional[datetime]:
        """Get the collection creation timestamp.

        Returns:
            Optional[datetime]: The timestamp when the collection was created, or None if not available.
        """
        return self._collection_instance.created_at

    @property
    def last_edited_at(self) -> Optional[datetime]:
        """Get the collection last edit timestamp.

        Returns:
            Optional[datetime]: The timestamp when the collection was last edited, or None if not available.
        """
        return self._collection_instance.last_edited_at

    @property
    def collection_type(self) -> ProjectCollectionType:
        """Get the type of the collection.

        Returns:
            ProjectCollectionType: The type of the collection.
        """
        return self._collection_instance.collection_type

    @property
    def project_hash(self) -> UUID:
        """Get the project hash of the collection.

        Returns:
            UUID: The project hash of the collection.
        """
        return self._collection_instance.project_uuid

    @staticmethod
    def _get_collection(
        project_client: EncordClientProject,
        ontology: Ontology,
        project_uuid: UUID,
        collection_uuid: UUID,
    ) -> "ProjectCollection":
        params = GetProjectCollectionParams(uuids=[collection_uuid])
        orm_items = list(
            project_client._api_client.get_paged_iterator(
                f"active/{project_uuid}/collections",
                params=params,
                result_type=OrmProjectCollection,
            )
        )
        if len(orm_items) > 0:
            return ProjectCollection(
                project_uuid=project_uuid,
                project_client=project_client,
                ontology=ontology,
                orm_collection=orm_items[0],
            )
        raise AuthorisationError("No collection found")

    @staticmethod
    def _list_collections(
        project_client: EncordClientProject,
        ontology: Ontology,
        project_uuid: UUID,
        collection_uuids: Union[List[UUID], None],
        page_size: Optional[int] = None,
    ) -> Iterator["ProjectCollection"]:
        params = GetProjectCollectionParams(projectHash=project_uuid, uuids=collection_uuids, pageSize=page_size)
        paged_collections = project_client._api_client.get_paged_iterator(
            f"active/{project_uuid}/collections",
            params=params,
            result_type=OrmProjectCollection,
        )
        for collection in paged_collections:
            yield ProjectCollection(
                project_uuid=project_uuid,
                project_client=project_client,
                ontology=ontology,
                orm_collection=collection,
            )

    @staticmethod
    def _delete_collection(client: ApiClient, project_uuid: UUID, collection_uuid: UUID) -> None:
        client.delete(
            f"active/{project_uuid}/collections/{collection_uuid}",
            params=None,
            result_type=None,
        )

    @staticmethod
    def _create_collection(
        client: ApiClient,
        project_uuid: UUID,
        name: str,
        description: str = "",
        collection_type: ProjectCollectionType = ProjectCollectionType.FRAME,
    ) -> UUID:
        params = CreateProjectCollectionParams(projectHash=project_uuid)
        payload = CreateProjectCollectionPayload(name=name, description=description, collection_type=collection_type)
        return client.post(
            f"active/{project_uuid}/collections",
            params=params,
            payload=payload,
            result_type=UUID,
        )

    def update_collection(self, name: Optional[str] = None, description: Optional[str] = None) -> None:
        """Update the collection's name and/or description.

        Args:
           name (Optional[str]): The new name for the collection.
           description (Optional[str]): The new description for the collection.
        """
        payload = UpdateCollectionPayload(name=name, description=description)
        self._client.patch(
            f"active/{self._project_uuid}/collections/{self.uuid}",
            params=None,
            payload=payload,
            result_type=None,
        )

    def list_frames(
        self,
        page_size: Optional[int] = None,
    ) -> Iterator[Tuple[LabelRowV2, List[ProjectDataCollectionInstance]]]:
        """List frames in the collection.

        Args:
            page_size (Optional[int]): The number of items to fetch per page.

        Returns:
            Iterator[Tuple[LabelRowV2, List[ProjectDataCollectionInstance]]]: An list of tuples containing label
            row and corresponding frame instances in the collection.
        """
        params = GetCollectionItemsParams(pageSize=page_size)
        paged_items = self._client.get_paged_iterator(
            f"active/{self._project_uuid}/collections/{self.uuid}/items",
            params=params,
            result_type=ProjectDataCollectionItemResponse,
        )

        for item in paged_items:
            yield (
                LabelRowV2(
                    label_row_metadata_dto_to_label_row_metadata(item.label_row_metadata),
                    self._project_client,
                    self._ontology,
                ),
                item.instances,
            )

    def list_annotations(
        self,
        page_size: Optional[int] = None,
    ) -> Iterator[Tuple[LabelRowV2, List[ProjectLabelCollectionInstance]]]:
        """List annotations in the collection.

        Args:
            page_size (Optional[int]): The number of items to fetch per page.

        Returns:
            Iterator[Tuple[LabelRowV2, List[ProjectLabelCollectionInstance]]]: An list of tuples containing label
            row and corresponding label instances in the collection.
        """
        params = GetCollectionItemsParams(pageSize=page_size)
        paged_items = self._client.get_paged_iterator(
            f"active/{self._project_uuid}/collections/{self.uuid}/items?getLabels=true",
            params=params,
            result_type=ProjectLabelCollectionItemResponse,
        )

        for item in paged_items:
            yield (
                LabelRowV2(
                    label_row_metadata_dto_to_label_row_metadata(item.label_row_metadata),
                    self._project_client,
                    self._ontology,
                ),
                item.instances,
            )

    def add_items(
        self, items: List[Union[ProjectDataCollectionItemRequest, ProjectLabelCollectionItemRequest]]
    ) -> ProjectCollectionBulkItemResponse:
        """Add data items to the collection.

        Args:
            items (Sequence[ProjectDataCollectionItemRequest | ProjectLabelCollectionItemRequest]): The list of data items to be added.

        Returns:
            ProjectCollectionBulkItemResponse: The response after adding items to the collection.
        """
        res = self._client.post(
            f"active/{self._project_uuid}/collections/{self.uuid}/add-items",
            params=None,
            payload=ProjectCollectionBulkItemRequest(items=items),
            result_type=ProjectCollectionBulkItemResponse,
        )
        return res

    def remove_items(
        self, items: List[Union[ProjectDataCollectionItemRequest, ProjectLabelCollectionItemRequest]]
    ) -> ProjectCollectionBulkItemResponse:
        """Remove data items from the collection.

        Args:
            items (Sequence[ProjectDataCollectionItemRequest | ProjectLabelCollectionItemRequest]): The list of data items to be removed.

        Returns:
            ProjectCollectionBulkItemResponse: The response after removing items from the collection.
        """
        res = self._client.post(
            f"active/{self._project_uuid}/collections/{self.uuid}/remove-items",
            params=None,
            payload=ProjectCollectionBulkItemRequest(items=items),
            result_type=ProjectCollectionBulkItemResponse,
        )
        return res

    def add_preset_items(self, filter_preset: Union[FilterPreset, UUID, str]) -> None:
        """Async operation to add storage items matching a filter preset to the collection.

        Args:
             filter_preset (Union[FilterPreset, UUID, str]): The filter preset or its UUID/ID used to filter items.
        """
        if isinstance(filter_preset, FilterPreset):
            preset_uuid = filter_preset.uuid
        elif isinstance(filter_preset, str):
            preset_uuid = UUID(filter_preset)
        else:
            preset_uuid = filter_preset
        self._client.post(
            f"active/{self._project_uuid}/collections/{self.uuid}/add-preset-items",
            params=None,
            payload=CollectionBulkPresetRequest(preset_uuid=preset_uuid),
            result_type=None,
        )
        log.info(
            f"Submitted request to add items matching filter_preset:{preset_uuid} to collection:{self.uuid}."
            f"It is an async operation and can take some time to complete."
        )

    def remove_preset_items(self, filter_preset: Union[FilterPreset, UUID, str]) -> None:
        """Async operation to remove storage items matching a filter preset from the collection.

        Args:
            filter_preset (Union[FilterPreset, UUID, str]): The filter preset or its UUID/ID used to filter items.
        """
        if isinstance(filter_preset, FilterPreset):
            preset_uuid = filter_preset.uuid
        elif isinstance(filter_preset, str):
            preset_uuid = UUID(filter_preset)
        else:
            preset_uuid = filter_preset
        self._client.post(
            f"active/{self._project_uuid}/collections/{self.uuid}/remove-preset-items",
            params=None,
            payload=CollectionBulkPresetRequest(preset_uuid=preset_uuid),
            result_type=None,
        )
        log.info(
            f"Submitted request to remove items matching filter_preset:{preset_uuid} from collection:{self.uuid}."
            f"It is an async operation and can take some time to complete."
        )



from encord.exceptions import (
    AnswerDictionaryError,
    AuthenticationError,
    AuthorisationError,
    CorruptedLabelError,
    DetectionRangeInvalidError,
    DuplicateSshKeyError,
    FeatureDoesNotExistError,
    FileSizeNotSupportedError,
    FileTypeNotSupportedError,
    GenericServerError,
    InvalidArgumentsError,
    InvalidDateFormatError,
    MethodNotAllowedError,
    ModelFeaturesInconsistentError,
    ModelWeightsInconsistentError,
    MultiLabelLimitError,
    OperationNotAllowed,
    ResourceExistsError,
    ResourceNotFoundError,
    SshKeyNotFound,
    UnknownException,
    UploadOperationNotSupportedError,
    WrongProjectTypeError,
)

# Error messages
AUTHENTICATION_ERROR = ["AUTHENTICATION_ERROR"]
AUTHORISATION_ERROR = ["AUTHORISATION_ERROR"]
RESOURCE_NOT_FOUND_ERROR = ["RESOURCE_NOT_FOUND_ERROR"]
METHOD_NOT_ALLOWED_ERROR = ["METHOD_NOT_ALLOWED_ERROR"]
UNKNOWN_ERROR = ["UNKNOWN_ERROR"]
OPERATION_NOT_ALLOWED_ERROR = ["OPERATION_NOT_ALLOWED"]
ANSWER_DICTIONARY_ERROR = ["ANSWER_DICTIONARY_ERROR"]
CORRUPTED_LABEL_ERROR = ["CORRUPTED_LABEL_ERROR"]
FILE_TYPE_NOT_SUPPORTED_ERROR = ["FILE_TYPE_NOT_SUPPORTED_ERROR"]
FILE_SIZE_NOT_SUPPORTED_ERROR = ["FILE_SIZE_NOT_SUPPORTED_ERROR"]
FEATURE_DOES_NOT_EXIST_ERROR = ["FEATURE_DOES_NOT_EXIST_ERROR"]
MODEL_WEIGHTS_INCONSISTENT_ERROR = ["MODEL_WEIGHTS_INCONSISTENT_ERROR"]
MODEL_FEATURES_INCONSISTENT_ERROR = ["MODEL_FEATURES_INCONSISTENT_ERROR"]
UPLOAD_OPERATION_NOT_SUPPORTED_ERROR = ["UPLOAD_OPERATION_NOT_SUPPORTED_ERROR"]
MUST_SET_DETECTION_RANGE_ERROR = ["MUST_SET_DETECTION_RANGE_ERROR"]
DETECTION_RANGE_INVALID_ERROR = ["DETECTION_RANGE_INVALID_ERROR"]
RESOURCE_EXISTS_ERROR = ["RESOURCE_EXISTS_ERROR"]
INVALID_DATE_FORMAT_ERROR = ["INVALID_DATE_FORMAT_ERROR"]
DUPLICATE_SSH_KEY_ERROR = ["DUPLICATE_SSH_KEY_ERROR"]
SSH_KEY_NOT_FOUND_ERROR = ["SSH_KEY_NOT_FOUND_ERROR"]
INVALID_ARGUMENTS_ERROR = ["INVALID_ARGUMENTS_ERROR"]
MULTI_LABEL_LIMIT_ERROR = ["MULTI_LABEL_LIMIT_ERROR"]
WRONG_PROJECT_TYPE_ERROR = ["WRONG_PROJECT_TYPE_ERROR"]


def check_error_response(response, context=None, payload=None):
    """Checks server response.
    Called if HTTP response status code is an error response.
    """
    if response == AUTHENTICATION_ERROR:
        raise AuthenticationError("You are not authenticated to access the Encord platform.", context=context)

    if response == AUTHORISATION_ERROR:
        raise AuthorisationError("You are not authorised to access this asset.", context=context)

    if response == RESOURCE_NOT_FOUND_ERROR:
        if payload:
            raise ResourceNotFoundError(payload, context=context)
        raise ResourceNotFoundError("The requested resource was not found in the database.", context=context)

    if response == METHOD_NOT_ALLOWED_ERROR:
        raise MethodNotAllowedError("HTTP method is not allowed.", context=context)

    if response == UNKNOWN_ERROR:
        raise UnknownException("An unknown error occurred.", context=context)

    if response == OPERATION_NOT_ALLOWED_ERROR:
        raise OperationNotAllowed("The read/write operation is not allowed by the API key.", context=context)

    if response == ANSWER_DICTIONARY_ERROR:
        raise AnswerDictionaryError(
            "An object or classification is missing in the answer dictionaries.", context=context
        )

    if response == CORRUPTED_LABEL_ERROR:
        message = "The label blurb is corrupted."

        if payload and isinstance(payload, str):
            message += "\n" + payload
        else:
            message += (
                " This could be due to the number of frame labels exceeding the number of frames in the labelled video."
            )
        raise CorruptedLabelError(
            message=message,
            context=context,
        )

    if response == FILE_TYPE_NOT_SUPPORTED_ERROR:
        raise FileTypeNotSupportedError(
            "Supported file types are: image/jpeg, image/png, video/webm, video/mp4.", context=context
        )

    if response == FILE_SIZE_NOT_SUPPORTED_ERROR:
        raise FileSizeNotSupportedError(
            "The combined size of the input files is larger than the supported limit", context=context
        )

    if response == FEATURE_DOES_NOT_EXIST_ERROR:
        raise FeatureDoesNotExistError("The passed feature does not exist in the project ontology.", context=context)

    if response == MODEL_WEIGHTS_INCONSISTENT_ERROR:
        raise ModelWeightsInconsistentError(
            "The passed model weights are incompatible with the selected model.", context=context
        )

    if response == MODEL_FEATURES_INCONSISTENT_ERROR:
        raise ModelFeaturesInconsistentError(
            "The passed features are incompatible with the selected model.", context=context
        )

    if response == UPLOAD_OPERATION_NOT_SUPPORTED_ERROR:
        raise UploadOperationNotSupportedError(
            "Uploading a file to an external (e.g. S3/GCP/Azure) dataset is not permitted.", context=context
        )

    if response == DETECTION_RANGE_INVALID_ERROR:
        raise DetectionRangeInvalidError(
            "The detection range is invalid (e.g. less than 0, or higher than num frames in the video)",
            context=context,
        )

    if response == INVALID_DATE_FORMAT_ERROR:
        raise InvalidDateFormatError("Invalid date format supplied as input", context=context)

    if response == RESOURCE_EXISTS_ERROR:
        raise ResourceExistsError(
            f"Trying to create a resource that already exists. Payload for this failure is: {payload}",
            context=context,
        )

    if response == DUPLICATE_SSH_KEY_ERROR:
        raise DuplicateSshKeyError(
            "The used SSH key exists more than once. Please create a unique user SSH key.", context=context
        )

    if response == SSH_KEY_NOT_FOUND_ERROR:
        raise SshKeyNotFound(
            "The used SSH key does not exist on the Encord platform. Please add this SSH key to your user profile.",
            context=context,
        )

    if response == INVALID_ARGUMENTS_ERROR:
        default_message = "Some of the arguments to the SDK function were invalid."
        message = default_message if payload is None else payload
        raise InvalidArgumentsError(message, context=context)

    if response == MULTI_LABEL_LIMIT_ERROR:
        maximum_labels_allowed = payload.get("maximum_labels_allowed", 100) if payload is not None else 100
        raise MultiLabelLimitError(
            f"Too many labels were requested. The limit is {maximum_labels_allowed}. Please reduce the amount "
            "of requested labels to stay under the reported limit.",
            maximum_labels_allowed=maximum_labels_allowed,
            context=context,
        )

    if response == WRONG_PROJECT_TYPE_ERROR:
        default_message = "Operation is not supported by the project type"
        message = default_message if payload is None else payload
        raise WrongProjectTypeError(message, context=context)

    payload_string = ""
    if payload:
        payload_string = f" with the following payload `{payload}`"
    raise GenericServerError(
        f"The Encord server has reported an error of type `{response}`{payload_string}. Please do not parse this error "
        "programmatically, instead please upgrade the SDK to the latest version to get the exact error.",
        context=context,
    )



from enum import Enum


class ObjectShape(Enum):
    POLYGON = "polygon"
    POLYLINE = "polyline"
    BOUNDING_BOX = "bounding_box"
    KEY_POINT = "point"
    SKELETON = "skeleton"
    ROTATABLE_BOUNDING_BOX = "rotatable_bounding_box"



try:
    import pycocotools
    import shapely
except ImportError as e:
    raise ImportError(
        "The 'pycocotools' and 'shapely' packages are required for the COCO export. "
        "Install them with: `pip install encord[coco]`"
    ) from e

import logging
from collections import defaultdict
from dataclasses import dataclass
from itertools import chain
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import numpy as np
from pycocotools import mask as cocomask
from pydantic import BaseModel
from shapely.geometry import Polygon

from encord.exceptions import EncordException
from encord.objects.attributes import Attribute
from encord.objects.common import PropertyType, Shape
from encord.objects.ontology_object import Object
from encord.objects.ontology_structure import OntologyStructure

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class DicomAnnotationData:
    dicom_instance_uid: str
    multiframe_frame_number: Optional[int]  # only present for multi-frame DICOMs
    file_uri: str
    width: int
    height: int


class CocoAnnotation(BaseModel):
    area: float
    bbox: Tuple[float, float, float, float]
    category_id: int
    id_: int
    image_id: int
    is_crowd: int
    segmentation: Union[List[List[float]], Dict[str, Any]]
    keypoints: Optional[List[float]] = None
    num_keypoints: Optional[int] = None
    track_id: Optional[int] = None
    encord_track_uuid: Optional[str] = None
    rotation: Optional[float] = None
    classifications: Optional[Dict[str, Any]] = None
    manual_annotation: Optional[bool] = None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "area": self.area,
            "bbox": list(self.bbox),
            "category_id": self.category_id,
            "image_id": self.image_id,
            "iscrowd": self.is_crowd,
            "segmentation": self.segmentation,
            "keypoints": self.keypoints,
            "num_keypoints": self.num_keypoints,
            "id": self.id_,
            "attributes": {
                k: v
                for k, v in {
                    "track_id": self.track_id,
                    "encord_track_uuid": self.encord_track_uuid,
                    "rotation": self.rotation,
                    "classifications": self.classifications,
                    "manual_annotation": self.manual_annotation,
                }.items()
                if v is not None
            },
        }


@dataclass
class Size:
    width: int
    height: int


def get_polygon_from_dict(
    polygon_dict: Dict[str, Any],
    w: int,
    h: int,
) -> List[Tuple[float, float]]:
    return [
        (
            polygon_dict[str(i)]["x"] * w,
            polygon_dict[str(i)]["y"] * h,
        )
        for i in range(len(polygon_dict))
    ]


class CocoExporter:
    """This class is intentionally designed in a modular fashion to facilitate extensibility.
    You are encouraged to subclass this exporter and modify any of the custom functions.
    While return types are provided for convenience, they can also be subclassed to suit your specific requirements.
    All functions that could be static are intentionally not made static, allowing you the option to access the `self`
    object at any time.
    """

    def __init__(
        self,
        labels_list: List[Dict[str, Any]],
        ontology: OntologyStructure,
        include_videos: bool = True,
    ) -> None:
        self._labels_list = labels_list
        self._ontology = ontology
        self._coco_json: Dict[str, Any] = {}
        self._current_annotation_id = 0
        self._object_hash_to_track_id_map: Dict[str, int] = {}
        self._coco_categories_id_to_ontology_object_map: Dict = {}  # TODO: do we need this?
        self._feature_hash_to_coco_category_id_map: Dict[str, int] = {}
        self._data_hash_to_image_id_map: Dict[Tuple[str, int], int] = {}
        """Map of (data_hash, frame_offset) to the image id"""

        self._feature_hash_to_attribute_map: Optional[Dict[str, Attribute]] = None
        self._id_and_object_hash_to_answers_map: Optional[Dict[Tuple[int, str], Dict]] = None
        self._include_videos = include_videos

    def export(self) -> Dict[str, Any]:
        self._coco_json["info"] = self.get_info()
        self._coco_json["categories"] = self.get_categories()
        self._coco_json["images"] = self.get_images()
        self._coco_json["annotations"] = [x.to_dict() for x in self.get_all_annotations()]

        return self._coco_json

    def get_info(self) -> Dict[str, Optional[str]]:
        return {
            "description": self.get_description(),
            "contributor": None,  # TODO: these fields also need a response
            "date_created": None,
            # TODO: there is something in the labels, alternatively can start to return more from the SDK
            "url": None,
            "version": None,
            "year": None,
        }

    def get_description(self) -> Optional[str]:
        if len(self._labels_list) == 0:
            res: Optional[str] = None
        else:
            res = str(self._labels_list[0]["data_title"])

        return res

    def get_categories(self) -> List[Dict[str, Any]]:
        """This does not translate classifications as they are not part of the Coco spec."""
        categories = []
        for object_ in self._ontology.objects:
            categories.append(self.get_category(object_))

        return categories

    def get_category(self, object_: Object) -> Dict[str, Any]:
        super_category = self.get_super_category(object_)
        ret: Dict[str, Any] = {
            "supercategory": super_category,
            "id": self.add_to_object_map_and_get_next_id(object_),
            "name": self.get_category_name(object_),
        }

        if super_category == "point":
            # TODO: we will have to do something similar for skeletons.
            ret["keypoints"] = "keypoint"
            ret["skeleton"] = []

        return ret

    def get_super_category(self, object_: Object) -> str:
        return object_.shape.value

    def add_to_object_map_and_get_next_id(self, object_: Object) -> int:
        id_ = len(self._coco_categories_id_to_ontology_object_map) + 1
        # Let the category id start at 1, not 0. Segmentation masks that use COCO annotations often create a bitmask
        # with this category id. The bitmask must be above 0. See this link:
        # https://cocodataset.org/#stuff-eval
        self._coco_categories_id_to_ontology_object_map[id_] = object_
        self._feature_hash_to_coco_category_id_map[object_.feature_node_hash] = id_

        return id_

    def get_category_name(self, object_: Object) -> str:
        return object_.name

    def get_images(self) -> List[Dict[str, Any]]:
        """All the data is in the specific label_row"""
        images = []

        for labels in self._labels_list:
            for data_unit in labels["data_units"].values():
                data_type = data_unit["data_type"]
                if "application/dicom" in data_type:
                    images.extend(self.get_dicom(data_unit))
                elif "video" not in data_type:
                    images.append(self.get_image(data_unit))
                else:
                    images.extend(self.get_video_images(data_unit))

        return images

    def get_dicom(self, data_unit: Dict) -> List:
        return [
            self._dicom_label_to_coco_image(
                int(key), data_unit["data_hash"], data_unit["width"], data_unit["height"], label
            )
            for key, label in data_unit["labels"].items()
        ]

    def get_image(self, data_unit: Dict[str, Any]) -> Dict[str, Any]:
        # TODO: we probably want a map of this image id to image hash in our DB, including the image_group hash.
        image_id = len(self._data_hash_to_image_id_map)
        data_hash = data_unit["data_hash"]
        self._data_hash_to_image_id_map[(data_hash, 0)] = image_id

        return {
            "coco_url": data_unit["data_link"],
            "id": image_id,
            "image_title": data_unit["data_title"],
            "file_name": f'images/{data_unit["data_hash"]}.{data_unit["data_title"].split(".")[-1]}',
            "height": data_unit["height"],
            "width": data_unit["width"],
        }

    def get_video_images(self, data_unit: Dict[str, Any]) -> List[Dict[str, Any]]:
        if not self._include_videos:
            return []

        video_title = data_unit["data_title"]
        data_hash = data_unit["data_hash"]

        images = []
        coco_url = data_unit["data_link"]
        height = data_unit["height"]
        width = data_unit["width"]

        for frame_num in data_unit["labels"].keys():
            images.append(
                self.get_video_image(
                    data_hash,
                    video_title,
                    coco_url,
                    height,
                    width,
                    int(frame_num),
                )
            )

        return images

    def _dicom_label_to_coco_image(
        self, frame: int, data_hash: str, series_width: int, series_height: int, dicom_label: Dict
    ) -> Dict[str, Any]:
        image_id = len(self._data_hash_to_image_id_map)
        # ideally this should be verify_arg, but currently we can't be sure that the metadata is on every frame
        metadata = dicom_label.get("metadata")
        self._data_hash_to_image_id_map[(data_hash, frame)] = image_id
        if metadata:
            file_name = f"dicom/{data_hash}/{metadata.dicom_instance_uid}"
            if metadata.multiframe_frame_number is not None:
                file_name += f"/{metadata.multiframe_frame_number}"

            return {
                "coco_url": metadata.file_uri,
                "id": image_id,
                "file_name": file_name,
                "height": metadata.height,
                "width": metadata.width,
            }
        else:
            return {
                "coco_url": "",
                "id": image_id,
                "file_name": f"dicom/{data_hash}/{frame}",
                "height": series_height,
                "width": series_width,
            }

    def get_video_image(
        self,
        data_hash: str,
        video_title: str,
        coco_url: str,
        height: int,
        width: int,
        frame_num: int,
    ) -> Dict[str, Any]:
        image_id = len(self._data_hash_to_image_id_map)
        self._data_hash_to_image_id_map[(data_hash, frame_num)] = image_id

        return {
            "coco_url": coco_url,
            "id": image_id,
            "video_title": video_title,
            "file_name": f"videos/{data_hash}/{frame_num}.jpg",
            "height": height,
            "width": width,
        }

    def get_all_annotations(self) -> List[CocoAnnotation]:
        annotations = []

        # TODO: need to make sure at least one image
        for labels in self._labels_list:
            object_answers = labels["object_answers"]
            object_actions = labels["object_actions"]

            for data_unit in labels["data_units"].values():
                data_hash = data_unit["data_hash"]

                if "video" in data_unit["data_type"]:
                    if not self._include_videos:
                        continue
                    for frame_num, frame_item in data_unit["labels"].items():
                        image_id = self.get_image_id(data_hash, int(frame_num))
                        objects = frame_item["objects"]
                        annotations.extend(
                            self.get_annotations(
                                objects,
                                image_id,
                                object_answers,
                                object_actions,
                            )
                        )

                elif "application/dicom" in data_unit["data_type"]:
                    # copy pasta:
                    for frame_num, frame_item in data_unit["labels"].items():
                        image_id = self.get_image_id(data_hash, int(frame_num))
                        objects = frame_item["objects"]
                        annotations.extend(
                            self.get_annotations(
                                objects,
                                image_id,
                                object_answers,
                                object_actions,
                            )
                        )

                else:
                    image_id = self.get_image_id(data_hash)
                    objects = data_unit["labels"].get("objects") or []
                    annotations.extend(
                        self.get_annotations(
                            objects,
                            image_id,
                            object_answers,
                            object_actions,
                        )
                    )

        return annotations

    def get_annotations(
        self,
        objects: List[Dict],
        image_id: int,
        object_answers: Dict,
        object_actions: Dict,
    ) -> List[CocoAnnotation]:
        annotations = []

        for object_ in objects:
            shape = object_["shape"]

            for image_data in self._coco_json["images"]:
                if image_data["id"] == image_id:
                    size = Size(width=image_data["width"], height=image_data["height"])

            if shape == Shape.BOUNDING_BOX.value:
                annotations.append(
                    self.get_bounding_box(
                        object_,
                        image_id,
                        size,
                        object_answers,
                        object_actions,
                    )
                )
            elif shape == Shape.ROTATABLE_BOUNDING_BOX.value:
                annotations.append(
                    self.get_rotatable_bounding_box(
                        object_,
                        image_id,
                        size,
                        object_answers,
                        object_actions,
                    )
                )
            elif shape == Shape.POLYGON.value:
                annotations.append(
                    self.get_polygon(
                        object_,
                        image_id,
                        size,
                        object_answers,
                        object_actions,
                    )
                )
            elif shape == Shape.POLYLINE.value:
                annotations.append(
                    self.get_polyline(
                        object_,
                        image_id,
                        size,
                        object_answers,
                        object_actions,
                    )
                )
            elif shape == Shape.BITMASK.value:
                annotations.append(
                    self.get_bitmask(
                        object_,
                        image_id,
                        size,
                        object_answers,
                        object_actions,
                    )
                )
            elif shape == Shape.POINT.value:
                annotations.append(
                    self.get_point(
                        object_,
                        image_id,
                        size,
                        object_answers,
                        object_actions,
                    )
                )
            elif shape == Shape.SKELETON.value:
                annotations.append(
                    self.get_skeleton(
                        object_,
                        image_id,
                        size,
                        object_answers,
                        object_actions,
                    )
                )

        return annotations

    def get_bounding_box(
        self,
        object_: Dict,
        image_id: int,
        size: Size,
        object_answers: Dict,
        object_actions: Dict,
    ) -> CocoAnnotation:
        x, y = (
            object_["boundingBox"]["x"] * size.width,
            object_["boundingBox"]["y"] * size.height,
        )
        w, h = (
            object_["boundingBox"]["w"] * size.width,
            object_["boundingBox"]["h"] * size.height,
        )
        area = w * h
        segmentation = [[x, y, x + w, y, x + w, y + h, x, y + h]]
        bbox = (x, y, w, h)
        category_id = self.get_category_id(object_)
        id_, is_crowd, track_id, encord_track_uuid, manual_annotation = self.get_coco_annotation_default_fields(object_)

        classifications = self.get_flat_classifications(object_, image_id, object_answers, object_actions)

        return CocoAnnotation(
            area=area,
            bbox=bbox,
            category_id=category_id,
            id_=id_,
            image_id=image_id,
            is_crowd=is_crowd,
            segmentation=segmentation,
            keypoints=None,
            num_keypoints=None,
            track_id=track_id,
            encord_track_uuid=encord_track_uuid,
            rotation=None,
            classifications=classifications,
            manual_annotation=manual_annotation,
        )

    def get_rotatable_bounding_box(
        self,
        object_: Dict,
        image_id: int,
        size: Size,
        object_answers: Dict,
        object_actions: Dict,
    ) -> CocoAnnotation:
        x, y = (
            object_["rotatableBoundingBox"]["x"] * size.width,
            object_["rotatableBoundingBox"]["y"] * size.height,
        )
        w, h = (
            object_["rotatableBoundingBox"]["w"] * size.width,
            object_["rotatableBoundingBox"]["h"] * size.height,
        )

        area = w * h
        segmentation = [[x, y, x + w, y, x + w, y + h, x, y + h]]
        bbox = (x, y, w, h)
        category_id = self.get_category_id(object_)
        id_, is_crowd, track_id, encord_track_uuid, manual_annotation = self.get_coco_annotation_default_fields(object_)

        rotation = object_["rotatableBoundingBox"]["theta"]

        classifications = self.get_flat_classifications(object_, image_id, object_answers, object_actions)

        return CocoAnnotation(
            area=area,
            bbox=bbox,
            category_id=category_id,
            id_=id_,
            image_id=image_id,
            is_crowd=is_crowd,
            segmentation=segmentation,
            keypoints=None,
            num_keypoints=None,
            track_id=track_id,
            encord_track_uuid=encord_track_uuid,
            rotation=rotation,
            classifications=classifications,
            manual_annotation=manual_annotation,
        )

    def get_polygon(
        self,
        object_: Dict,
        image_id: int,
        size: Size,
        object_answers: Dict,
        object_actions: Dict,
    ) -> CocoAnnotation:
        polygon = get_polygon_from_dict(object_["polygon"], size.width, size.height)
        segmentation = [list(chain(*polygon))]
        _polygon = Polygon(polygon)
        area: float = _polygon.area
        x, y, x_max, y_max = _polygon.bounds
        w, h = x_max - x, y_max - y

        bbox = (x, y, w, h)
        category_id = self.get_category_id(object_)
        id_, is_crowd, track_id, encord_track_uuid, manual_annotation = self.get_coco_annotation_default_fields(object_)

        classifications = self.get_flat_classifications(
            object_,
            image_id,
            object_answers,
            object_actions,
        )

        return CocoAnnotation(
            area=area,
            bbox=bbox,
            category_id=category_id,
            id_=id_,
            image_id=image_id,
            is_crowd=is_crowd,
            segmentation=segmentation,
            keypoints=None,
            num_keypoints=None,
            track_id=track_id,
            encord_track_uuid=encord_track_uuid,
            rotation=None,
            classifications=classifications,
            manual_annotation=manual_annotation,
        )

    def get_polyline(
        self,
        object_: Dict,
        image_id: int,
        size: Size,
        object_answers: Dict,
        object_actions: Dict,
    ) -> CocoAnnotation:
        """Polylines are technically not supported in COCO, but here we use a trick to allow a representation."""
        polygon = get_polygon_from_dict(object_["polyline"], size.width, size.height)
        polyline_coordinate = self.join_polyline_from_polygon(list(chain(*polygon)))
        segmentation = [polyline_coordinate]
        area = 0
        bbox = self.get_bbox_for_polyline(polygon)
        category_id = self.get_category_id(object_)
        id_, is_crowd, track_id, encord_track_uuid, manual_annotation = self.get_coco_annotation_default_fields(object_)

        classifications = self.get_flat_classifications(object_, image_id, object_answers, object_actions)

        return CocoAnnotation(
            area=area,
            bbox=bbox,
            category_id=category_id,
            id_=id_,
            image_id=image_id,
            is_crowd=is_crowd,
            segmentation=segmentation,
            keypoints=None,
            num_keypoints=None,
            track_id=track_id,
            encord_track_uuid=encord_track_uuid,
            rotation=None,
            classifications=classifications,
            manual_annotation=manual_annotation,
        )

    def get_bbox_for_polyline(
        self,
        polygon: List[Tuple[float, float]],
    ) -> Tuple[float, float, float, float]:
        if len(polygon) == 2:
            # We have the edge case of a single edge polygon.
            first_point = polygon[0]
            second_point = polygon[1]
            x = float(min(first_point[0], second_point[0]))
            y = float(min(first_point[1], second_point[1]))
            w = float(abs(first_point[0] - second_point[0]))
            h = float(abs(first_point[1] - second_point[1]))
            return x, y, w, h
        else:
            shapely_polygon = Polygon(polygon)
            x, y, x_max, y_max = shapely_polygon.bounds  # type: ignore[attr-defined]
            w, h = x_max - x, y_max - y

            return x, y, w, h

    @staticmethod
    def join_polyline_from_polygon(
        polygon: List[float],
    ) -> List[float]:
        """Essentially a trick to represent a polyline in coco. We pretend for this to be a polygon and join every
        coordinate from the end back to the beginning, so it will essentially be an area-less polygon.
        This function technically changes the input polygon in place.
        """
        if len(polygon) % 2 != 0:
            raise RuntimeError("The polygon has an unaccepted shape.")

        idx = len(polygon) - 2
        while idx >= 0:
            y_coordinate = polygon[idx]
            x_coordinate = polygon[idx + 1]
            polygon.append(y_coordinate)
            polygon.append(x_coordinate)
            idx -= 2

        return polygon

    def get_bitmask(
        self,
        object_: Dict,
        image_id: int,
        size: Size,
        object_answers: Dict,
        object_actions: Dict,
    ) -> CocoAnnotation:
        bitmask = object_["bitmask"]
        # Note: It's essential to transpose the input and output coordinates when using pycocotools' functions,
        # in order to address the distinction between treating masks as C-contiguous (row-major order, Encord's
        # implementation) versus pycocotools' expectation of Fortran-contiguous (column-major order, COCO's API
        # implementation) data.

        # Obtain the COCO compatible RLE string (convert from row-major to column-major order)
        transposed_segmentation = dict(counts=bitmask["rleString"], size=[bitmask["width"], bitmask["height"]])
        mask = np.asfortranarray(cocomask.decode(transposed_segmentation).T)
        segmentation = cocomask.encode(mask)
        bbox = tuple(cocomask.toBbox(segmentation))
        area = float(cocomask.area(segmentation))
        # Convert RLE string from bytes (which is not a JSON serializable type) to a string format
        segmentation["counts"] = segmentation["counts"].decode("ascii")

        category_id = self.get_category_id(object_)
        id_, _, track_id, encord_track_uuid, manual_annotation = self.get_coco_annotation_default_fields(object_)
        is_crowd = 1

        classifications: Optional[Dict] = self.get_flat_classifications(
            object_, image_id, object_answers, object_actions
        )

        return CocoAnnotation(
            area=area,
            bbox=bbox,
            category_id=category_id,
            id_=id_,
            image_id=image_id,
            is_crowd=is_crowd,
            segmentation=segmentation,
            keypoints=None,
            num_keypoints=None,
            track_id=track_id,
            encord_track_uuid=encord_track_uuid,
            rotation=None,
            classifications=classifications,
            manual_annotation=manual_annotation,
        )

    def get_point(
        self,
        object_: Dict,
        image_id: int,
        size: Size,
        object_answers: Dict,
        object_actions: Dict,
    ) -> CocoAnnotation:
        x, y = (
            object_["point"]["0"]["x"] * size.width,
            object_["point"]["0"]["y"] * size.height,
        )
        w, h = 0, 0
        area = 0
        segmentation = [[x, y]]
        keypoints = [x, y, 2]
        num_keypoints = 1

        bbox = (x, y, w, h)
        category_id = self.get_category_id(object_)
        id_, is_crowd, track_id, encord_track_uuid, manual_annotation = self.get_coco_annotation_default_fields(object_)

        classifications: Optional[Dict] = self.get_flat_classifications(
            object_, image_id, object_answers, object_actions
        )

        return CocoAnnotation(
            area=area,
            bbox=bbox,
            category_id=category_id,
            id_=id_,
            image_id=image_id,
            is_crowd=is_crowd,
            segmentation=segmentation,
            keypoints=keypoints,
            num_keypoints=num_keypoints,
            track_id=track_id,
            encord_track_uuid=encord_track_uuid,
            rotation=None,
            classifications=classifications,
            manual_annotation=manual_annotation,
        )

    def get_skeleton(
        self,
        object_: Dict,
        image_id: int,
        size: Size,
        object_answers: Dict,
        object_actions: Dict,
    ) -> CocoAnnotation:
        area = 0
        segmentation: List = []
        keypoints = []

        for point in object_["skeleton"].values():
            keypoints += [
                point["x"] * size.width,
                point["y"] * size.height,
                2,
            ]

        num_keypoints = len(keypoints) // 3
        xs, ys = (
            keypoints[::3],
            keypoints[1::3],
        )
        x, y, x_max, y_max = min(xs), min(ys), max(xs), max(ys)
        w, h = x_max - x, y_max - y

        bbox = (x, y, w, h)
        category_id = self.get_category_id(object_)
        id_, is_crowd, track_id, encord_track_uuid, manual_annotation = self.get_coco_annotation_default_fields(object_)

        classifications: Optional[Dict] = self.get_flat_classifications(
            object_, image_id, object_answers, object_actions
        )

        return CocoAnnotation(
            area=area,
            bbox=bbox,
            category_id=category_id,
            id_=id_,
            image_id=image_id,
            is_crowd=is_crowd,
            segmentation=segmentation,
            keypoints=keypoints,
            num_keypoints=num_keypoints,
            track_id=track_id,
            encord_track_uuid=encord_track_uuid,
            rotation=None,
            classifications=classifications,
            manual_annotation=manual_annotation,
        )

    def get_flat_classifications(
        self, object_: Dict, image_id: int, object_answers: Dict, object_actions: Dict
    ) -> Dict[str, Any]:
        object_hash = object_["objectHash"]
        feature_hash = object_["featureHash"]

        feature_hash_to_attribute_map: Dict[str, Attribute] = self.get_feature_hash_to_flat_object_attribute_map()
        id_and_object_hash_to_answers_map = self.get_id_and_object_hash_to_answers_map(object_actions)

        classifications = self.get_flat_static_classifications(
            object_hash, feature_hash, object_answers, feature_hash_to_attribute_map
        )
        dynamic_classifications = self.get_flat_dynamic_classifications(
            object_hash,
            feature_hash,
            image_id,
            id_and_object_hash_to_answers_map,
        )
        classifications.update(dynamic_classifications)

        return classifications

    def get_feature_hash_to_flat_object_attribute_map(self) -> Dict[str, Attribute]:
        if self._feature_hash_to_attribute_map is not None:
            return self._feature_hash_to_attribute_map

        ret: Dict[str, Attribute] = {}

        for object_ in self._ontology.objects:
            for attribute in object_.attributes:
                ret[attribute.feature_node_hash] = attribute

        self._feature_hash_to_attribute_map = ret

        return ret

    def get_flat_static_classifications(
        self,
        object_hash: str,
        object_feature_hash: str,
        object_answers: Dict[str, Any],
        feature_hash_to_attribute_map: Dict[str, Attribute],
    ) -> Dict[str, Any]:
        ret = {}
        classifications = object_answers[object_hash]["classifications"]
        for classification in classifications:
            feature_hash = classification["featureHash"]
            if feature_hash not in feature_hash_to_attribute_map:
                # This will be a deeply nested attribute
                continue

            attribute = feature_hash_to_attribute_map[feature_hash]
            answers = classification["answers"]

            if attribute.get_property_type() == PropertyType.TEXT:
                ret.update(self.get_text_answer(attribute, answers))
            elif attribute.get_property_type() == PropertyType.RADIO:
                ret.update(self.get_radio_answer(attribute, answers))
            elif attribute.get_property_type() == PropertyType.CHECKLIST:
                ret.update(self.get_checklist_answer(attribute, answers))

        self.add_unselected_attributes(object_feature_hash, ret, match_dynamic_attributes=False)

        return ret

    def get_id_and_object_hash_to_answers_map(
        self,
        object_actions: Dict[str, Any],
    ) -> Dict[Tuple[int, str], Dict]:
        if self._id_and_object_hash_to_answers_map is not None:
            return self._id_and_object_hash_to_answers_map

        ret: Dict[Tuple[int, str], Dict[str, Any]] = defaultdict(Dict)
        feature_hash_to_attribute_map = self.get_feature_hash_to_flat_object_attribute_map()
        for object_hash, payload in object_actions.items():
            for action in payload["actions"]:
                feature_hash = action["featureHash"]
                if feature_hash not in feature_hash_to_attribute_map:
                    # This will be a deeply nested attribute
                    continue

                attribute = feature_hash_to_attribute_map[feature_hash]
                answers = action["answers"]
                answers_dict: Dict[str, Any] = {}

                if attribute.get_property_type() == PropertyType.TEXT:
                    answers_dict.update(self.get_text_answer(attribute, answers))
                elif attribute.get_property_type() == PropertyType.RADIO:
                    answers_dict.update(self.get_radio_answer(attribute, answers))
                elif attribute.get_property_type() == PropertyType.CHECKLIST:
                    answers_dict.update(self.get_checklist_answer(attribute, answers))

                for sub_range in action["range"]:
                    for i in range(sub_range[0], sub_range[1] + 1):
                        ret[(i, object_hash)].update(answers_dict)

        self._id_and_object_hash_to_answers_map = ret

        return ret

    def get_flat_dynamic_classifications(
        self,
        object_hash: str,
        feature_hash: str,
        image_id: int,
        id_and_object_hash_to_answers_map: Dict[Tuple[int, str], Dict[str, Any]],
    ) -> Dict[str, Any]:
        ret = {}
        id_and_object_hash = (image_id, object_hash)

        if id_and_object_hash in id_and_object_hash_to_answers_map:
            ret = id_and_object_hash_to_answers_map[(image_id, object_hash)]

        self.add_unselected_attributes(feature_hash, ret, match_dynamic_attributes=True)

        return ret

    def add_unselected_attributes(
        self, feature_hash: str, attributes_dict: Dict[str, Optional[bool]], match_dynamic_attributes: bool
    ) -> None:
        """Attributes which have never been selected will not show up in the actions map. They will need to be
        added separately. NOTE: this assumes uniqueness of features. Quite an edge case but if it ever comes
        up it needs to be solved somewhere here.
        """
        # TODO: This could be improved
        all_attributes = self.get_attributes_for_feature_hash(feature_hash)
        for attribute in all_attributes:
            is_matching_attribute = attribute.dynamic == match_dynamic_attributes
            if is_matching_attribute:
                if attribute.get_property_type() == PropertyType.CHECKLIST:
                    for option in attribute.options:  # type: ignore[union-attr]
                        if option.label not in attributes_dict:
                            # We need to add the default of False.
                            attributes_dict[option.label] = False
                else:
                    if attribute.name not in attributes_dict:
                        attributes_dict[attribute.name] = None

    def get_attributes_for_feature_hash(self, feature_hash: str) -> List[Attribute]:
        ret = []

        for object_ in self._ontology.objects:
            if object_.feature_node_hash == feature_hash:
                for attribute in object_.attributes:
                    ret.append(attribute)
                break

        return ret

    def get_radio_answer(self, attribute: Attribute, answers: List[Dict[str, str]]) -> Dict[str, str]:
        answer = answers[0]  # radios only have one answer by definition
        return {attribute.name: answer["name"]}

    def get_checklist_answer(self, attribute: Attribute, answers: List[Dict[str, Any]]) -> Dict[str, bool]:
        ret: Dict[str, bool] = {}
        found_checklist_answers: Set[str] = set()

        for answer in answers:
            found_checklist_answers.add(answer["name"])

        for option in attribute.options:  # type: ignore[union-attr]
            label = option.label
            ret[label] = label in found_checklist_answers

        return ret

    def get_text_answer(self, attribute: Attribute, answers: str) -> Dict[str, Any]:
        return {attribute.name: answers}

    def get_category_id(self, object_: Dict[str, Any]) -> int:
        feature_hash = object_["featureHash"]
        try:
            return self._feature_hash_to_coco_category_id_map[feature_hash]
        except KeyError:
            raise EncordException(
                f"The feature_hash `{feature_hash}` was not found in the provided ontology. Please "
                f"ensure that the ontology matches the labels provided."
            ) from None

    def get_coco_annotation_default_fields(
        self, object_: Dict[str, Any]
    ) -> Tuple[
        int,
        int,
        Optional[int],
        Optional[str],
        Optional[bool],
    ]:
        id_ = self.next_annotation_id()
        is_crowd = 0

        track_id = self.get_and_set_track_id(object_hash=object_["objectHash"])
        encord_track_uuid = object_["objectHash"]
        manual_annotation = object_["manualAnnotation"]

        return id_, is_crowd, track_id, encord_track_uuid, manual_annotation

    def next_annotation_id(self) -> int:
        next_ = self._current_annotation_id
        self._current_annotation_id += 1

        return next_

    def get_and_set_track_id(self, object_hash: str) -> int:
        if object_hash in self._object_hash_to_track_id_map:
            return self._object_hash_to_track_id_map[object_hash]
        else:
            next_track_id = len(self._object_hash_to_track_id_map)
            self._object_hash_to_track_id_map[object_hash] = next_track_id
            return next_track_id

    def get_image_id(self, data_hash: str, frame_num: int = 0) -> int:
        return self._data_hash_to_image_id_map[(data_hash, frame_num)]



import logging
from typing import (
    Dict,
    Optional,
)

from encord.objects.common import Shape
from encord.objects.ontology_labels_impl import LabelRowV2
from encord.objects.ontology_object import Object
from encord.project import Project
from encord.utilities.coco.datastructure import (
    CategoryID,
    CocoPolygon,
    CocoRLE,
    CocoRootModel,
    FrameIndex,
    ImageID,
)

logger = logging.getLogger()


def build_category_id_to_encord_ontology_object_map(
    project: Project,
    category_id_to_feature_hash: Dict[CategoryID, str],
) -> Dict[CategoryID, Object]:
    map_category_to_encord_object: Dict[CategoryID, Object] = {}
    for id, feature_hash in category_id_to_feature_hash.items():
        object_ = project.ontology_structure.get_child_by_hash(feature_node_hash=feature_hash, type_=Object)
        map_category_to_encord_object[id] = object_

    return map_category_to_encord_object


def initialise_label_rows(
    project: Project,
    image_id_to_frame_index: Dict[ImageID, FrameIndex],
    branch_name: Optional[str] = None,
) -> Dict[str, LabelRowV2]:
    data_hashes = list({frame_index.data_hash for frame_index in image_id_to_frame_index.values()})
    label_rows = project.list_label_rows_v2(data_hashes=data_hashes, branch_name=branch_name)
    with project.create_bundle() as bundle:
        for lr in label_rows:
            lr.initialise_labels(bundle=bundle)
    return {lr.data_hash: lr for lr in label_rows}


def import_coco_labels(
    project: Project,
    coco: CocoRootModel,
    category_id_to_feature_hash: Dict[CategoryID, str],
    image_id_to_frame_index: Dict[ImageID, FrameIndex],
    branch_name: Optional[str] = None,
) -> None:
    label_rows = initialise_label_rows(project, image_id_to_frame_index, branch_name=branch_name)
    category_id_to_objects = build_category_id_to_encord_ontology_object_map(project, category_id_to_feature_hash)
    coco_image_lookup = {i.id: i for i in coco.images}

    for i, annotation in enumerate(coco.annotations):
        frame_idx = image_id_to_frame_index.get(annotation.image_id)
        if frame_idx is None:
            # TODO not clear how to propagate errors
            logger.warning(
                f'Image id `{annotation.image_id}` from `labels_dict["annotations"][{i}]` could not be matched with the provided `image_id_to_frame_index`. Skipping this annotation.'
            )
            continue

        label_row = label_rows.get(frame_idx.data_hash)
        if label_row is None:
            logger.warning(
                f"Data hash `{frame_idx.data_hash}` from `image_id_to_frame_index` could not be matched with a data hash in the provided `project`. Skipping annotation ad index {i}."
            )
            # TODO not clear how to propagate errors
            continue

        ont_obj = category_id_to_objects.get(annotation.category_id)
        if ont_obj is None:
            logger.warning(
                f'Category ID {annotation.category_id} from `labels_dict["annotations"][{i}]` could not be matched with the provided `category_id_to_feature_hash`. Skipping this annotation.'
            )
            # TODO not clear how to propagate errors
            continue

        coco_image = coco_image_lookup.get(annotation.image_id)
        if coco_image is None:
            raise ValueError(
                f"The provided coco annotation dictionary have annotations with `image_id`s that do not match any image ids in the provided `images` list. Couldn't find image id {annotation.image_id}."
            )

        segmentation = annotation.segmentation
        width, height = coco_image.width, coco_image.height
        ins = ont_obj.create_instance()
        if ont_obj.shape == Shape.BOUNDING_BOX:
            ins.set_for_frames(
                coordinates=annotation.bbox.to_encord(img_w=width, img_h=height),
                frames=frame_idx.frame,
            )
            label_row.add_object_instance(ins)
        elif ont_obj.shape == Shape.BITMASK:
            if not isinstance(segmentation, CocoRLE):
                raise ValueError(
                    f"A mismatch between the format in the `labels_dict` and the expected shape in the ontology object with feature node hash {ont_obj.feature_node_hash} was detected for annotation id {annotation.id}. Expected format was an RLE."
                )
            ins.set_for_frames(coordinates=segmentation.to_encord(), frames=frame_idx.frame)
            label_row.add_object_instance(ins)
        elif ont_obj.shape == Shape.POLYGON:
            if not isinstance(segmentation, CocoPolygon) or (
                isinstance(segmentation, list) and isinstance(segmentation[0], CocoPolygon)
            ):
                raise ValueError(
                    f"A mismatch between the format in the `labels_dict` and the expected shape in the ontology object with feature node hash {ont_obj.feature_node_hash} was detected for annotation id {annotation.id}. Expected format was a polygon or a list of polygons. {segmentation}"
                )
            if isinstance(segmentation, CocoPolygon):
                segmentation = [segmentation]

            for poly in segmentation:
                ins.set_for_frames(
                    coordinates=poly.to_encord(img_w=width, img_h=height),
                    frames=frame_idx.frame,
                )
                label_row.add_object_instance(ins)
                ins = ont_obj.create_instance()
        else:
            raise ValueError(f"Ontology objects of shape {ont_obj.shape} are not supported for coco import")

    with project.create_bundle() as bundle:
        for label_row in label_rows.values():
            label_row.save(bundle=bundle)



from dataclasses import dataclass


@dataclass(frozen=True)
class ProjectWithUserRole:
    """This is a helper class denoting the relationship between the current user an a project"""

    user_role: int
    project: dict



from __future__ import annotations

import itertools
from dataclasses import dataclass
from typing import List, Optional, Union, cast
from uuid import UUID

import encord.orm.storage
from encord import EncordUserClient
from encord.http.bundle import Bundle, bundled_operation
from encord.http.v2.api_client import ApiClient
from encord.orm.base_dto import BaseDTO
from encord.orm.cloud_integration import CloudIntegration
from encord.storage import StorageItem

STORAGE_CLOUD_DATA_MIGRATION_BUNDLE_LIMIT = 1000


def update_storage_item_cloud_info(
    user_client: EncordUserClient,
    item: Union[StorageItem, UUID, str],
    new_url: Optional[str] = None,
    new_cloud_integration: Optional[Union[CloudIntegration, str, UUID]] = None,
    from_cloud_integration: Optional[Union[CloudIntegration, str, UUID]] = None,
    verify_access: bool = True,
    skip_missing: bool = False,
    bundle: Optional[Bundle] = None,
) -> None:
    """Update the cloud storage information of a storage item. This can be used to avoid re-importing data when the cloud
    configuration changes (and thus the data is accessible through a different cloud integration), or if the data has
    been moved to a different location (and thus the URL has changed).

    :param user_client: The user client to use.
    :param item: The item to update. Can be either a StorageItem instance, the item's UUID, or the URL of the item.
    :param new_url: The new URL to set. URL will be left unchanged if `None` is passed.
    :param new_cloud_integration: The new cloud integration to set. Cloud integration will be left unchanged if `None`
        is passed.
    :param from_cloud_integration: The cloud integration to update from. Acts as a check: no update will be performed if
        cloud integration of the item does not match this value. If `None`, the check will be skipped.
    :param verify_access: Whether to verify access to the item.
    :param skip_missing: if true, no error will be raised if the item is not found. Otherwise, the operation (and
        the operations in the same bundle) will be cancelled and an error will be raised.
    :param bundle: The optional :class:`encord.http.bundle.Bundle` instance used to group updates into bulk calls.
    """
    if not bundle:
        with Bundle() as bundle:
            update_storage_item_cloud_info(
                user_client=user_client,
                item=item,
                new_url=new_url,
                new_cloud_integration=new_cloud_integration,
                from_cloud_integration=from_cloud_integration,
                verify_access=verify_access,
                skip_missing=skip_missing,
                bundle=bundle,
            )
            return

    item_uuid: Optional[UUID] = None
    item_url: Optional[str] = None
    if isinstance(item, StorageItem):
        item_uuid = item.uuid
    elif isinstance(item, UUID):
        item_uuid = item
    else:  # it's a string
        try:
            item_uuid = UUID(item)
        except ValueError:
            item_url = item

    if not item_uuid and not item_url:
        raise ValueError("item must be either a StorageItem instance, a UUID, or a URL")

    if not new_url and not new_cloud_integration:
        raise ValueError("At least one of `new_url` or `new_cloud_integration` must be provided")

    new_cloud_integration_hash: Optional[UUID] = None

    if isinstance(new_cloud_integration, CloudIntegration):
        new_cloud_integration_hash = UUID(new_cloud_integration.id)
    elif isinstance(new_cloud_integration, UUID):
        new_cloud_integration_hash = new_cloud_integration
    elif isinstance(new_cloud_integration, str):
        new_cloud_integration_hash = UUID(new_cloud_integration)  # let the exception propagate

    from_cloud_integration_hash: Optional[UUID] = None

    if isinstance(from_cloud_integration, CloudIntegration):
        from_cloud_integration_hash = UUID(from_cloud_integration.id)
    elif isinstance(from_cloud_integration, UUID):
        from_cloud_integration_hash = from_cloud_integration
    elif isinstance(from_cloud_integration, str):
        from_cloud_integration_hash = UUID(from_cloud_integration)  # let the exception propagate

    bundled_operation(
        bundle,
        operation=user_client._api_client.get_bound_operation(_migrate_items_bundle),
        payload=BundledStorageMigrationPayload(
            item_migrations=[
                MigrateSingleItemPayload(
                    item_uuid=item_uuid,
                    item_url=item_url,
                    new_url=new_url,
                    new_cloud_integration=new_cloud_integration_hash,
                    from_cloud_integration=from_cloud_integration_hash,
                    verify_access=verify_access,
                    skip_missing=skip_missing,
                ),
            ],
        ),
        limit=STORAGE_CLOUD_DATA_MIGRATION_BUNDLE_LIMIT,
    )


class MigrateSingleItemPayload(BaseDTO):
    item_uuid: Optional[UUID]
    item_url: Optional[str]
    new_url: Optional[str]
    new_cloud_integration: Optional[UUID]
    from_cloud_integration: Optional[UUID]
    verify_access: bool
    skip_missing: bool

    def payload_key(self) -> str:
        return "::".join(
            str(part)
            for part in [
                "UUID" if self.item_uuid else "URL",
                self.new_cloud_integration,
                self.from_cloud_integration,
                self.verify_access,
                self.skip_missing,
            ]
        )


@dataclass
class BundledStorageMigrationPayload:
    item_migrations: List[MigrateSingleItemPayload]

    def add(self, other: BundledStorageMigrationPayload) -> BundledStorageMigrationPayload:
        self.item_migrations.extend(other.item_migrations)
        return self


def _migrate_items_bundle(
    api_client: ApiClient,
    *,
    item_migrations: List[MigrateSingleItemPayload],
) -> None:
    for _, single_call_items in itertools.groupby(
        sorted(((p.payload_key(), p) for p in item_migrations), key=lambda pp: pp[0]),
        key=lambda pp: pp[0],
    ):
        single_call_items_list = [pp[1] for pp in single_call_items]
        if not single_call_items_list:
            continue
        if single_call_items_list[0].item_uuid:
            items_map = {cast(UUID, p.item_uuid): p.new_url for p in single_call_items_list}
            urls_map = {}
        else:
            items_map = {}
            urls_map = {cast(str, p.item_url): p.new_url for p in single_call_items_list}

        api_client.post(
            path="/storage/items/migrate-storage",
            params=None,
            payload=encord.orm.storage.StorageItemsMigratePayload(
                urls_map=urls_map,
                items_map=items_map,
                from_integration_hash=single_call_items_list[0].from_cloud_integration,
                to_integration_hash=single_call_items_list[0].new_cloud_integration,
                validate_access=single_call_items_list[0].verify_access,
                skip_missing=single_call_items_list[0].skip_missing,
            ),
            result_type=None,
        )



"""---
title: "Annotation Stage"
slug: "sdk-ref-stage-annotation"
hidden: false
metadata:
  title: "Annotation Stage"
  description: "Encord SDK Annotation Stage."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from enum import Enum
from typing import Iterable, List, Literal, Optional, Union
from uuid import UUID

from encord.common.utils import ensure_list, ensure_uuid_list
from encord.http.bundle import Bundle
from encord.orm.workflow import WorkflowStageType
from encord.workflow.common import TasksQueryParams, WorkflowAction, WorkflowStageBase, WorkflowTask


class AnnotationTaskStatus(str, Enum):
    NEW = "NEW"
    ASSIGNED = "ASSIGNED"
    RELEASED = "RELEASED"
    SKIPPED = "SKIPPED"
    REOPENED = "REOPENED"
    COMPLETED = "COMPLETED"


class _AnnotationTasksQueryParams(TasksQueryParams):
    user_emails: Optional[List[str]] = None
    data_hashes: Optional[List[UUID]] = None
    dataset_hashes: Optional[List[UUID]] = None
    data_title_contains: Optional[str] = None
    statuses: Optional[List[AnnotationTaskStatus]] = None


class AnnotationStage(WorkflowStageBase):
    stage_type: Literal[WorkflowStageType.ANNOTATION] = WorkflowStageType.ANNOTATION

    """
    An Annotate stage in a non-Consensus Project.

    You can use this stage in Consensus and non-Consensus Projects.
    """

    def get_tasks(
        self,
        *,
        assignee: Union[List[str], str, None] = None,
        data_hash: Union[List[UUID], UUID, List[str], str, None] = None,
        dataset_hash: Union[List[UUID], UUID, List[str], str, None] = None,
        data_title: Optional[str] = None,
        status: Union[AnnotationTaskStatus, List[AnnotationTaskStatus], None] = None,
    ) -> Iterable[AnnotationTask]:
        """Retrieves tasks for the AnnotationStage.

        **Parameters**

        - `assignee` (Union[List[str], str, None]): A list of user emails or a single user email to filter tasks by assignee.
        - `data_hash` (Union[List[UUID], UUID, List[str], str, None]): A list of data unit UUIDs or a single data unit UUID to filter tasks by data hash.
        - `dataset_hash` (Union[List[UUID], UUID, List[str], str, None]): A list of dataset UUIDs or a single dataset UUID to filter tasks by dataset hash.
        - `data_title` (Optional[str]): A string to filter tasks by data title.
        - `status` (Optional[AnnotationTaskStatus | List[AnnotationTaskStatus]]): A status or a list of statuses to filter tasks by their status.

        **Returns**

        An iterable of `AnnotationTask` instances from both non-Consensus and Consensus Projects.
        """
        params = _AnnotationTasksQueryParams(
            user_emails=ensure_list(assignee),
            data_hashes=ensure_uuid_list(data_hash),
            dataset_hashes=ensure_uuid_list(dataset_hash),
            data_title_contains=data_title,
            statuses=ensure_list(status),
        )

        for task in self._workflow_client.get_tasks(self.uuid, params, type_=AnnotationTask):
            task._stage_uuid = self.uuid
            task._workflow_client = self._workflow_client
            yield task


class _ActionSubmit(WorkflowAction):
    action: Literal["SUBMIT"] = "SUBMIT"
    resolve_label_reviews: bool = True
    assignee: Optional[str] = None
    retain_assignee: bool = False


class _ActionAssign(WorkflowAction):
    action: Literal["ASSIGN"] = "ASSIGN"
    assignee: str


class _ActionRelease(WorkflowAction):
    action: Literal["RELEASE"] = "RELEASE"


class AnnotationTask(WorkflowTask):
    status: AnnotationTaskStatus
    data_hash: UUID
    data_title: str
    label_branch_name: str
    assignee: Optional[str]

    """
    Represents a task in a non-Consensus Annotate stage.

    **Attributes**

    - `status` (AnnotationTaskStatus): The current status of the task.
    - `data_hash` (UUID): Unique ID for the data unit.
    - `data_title` (str): Name of the data unit.
    - `label_branch_name` (str): Name of the label branch.
    - `assignee` (Optional[str]): User assigned to the task.

    **Allowed actions**

    - `submit`: Submits a task for review.
    - `assign`: Assigns a task to a user.
    - `release`: Releases a task from the current user.
    """

    def submit(
        self,
        *,
        assignee: Optional[str] = None,
        retain_assignee: bool = False,
        bundle: Optional[Bundle] = None,
    ) -> None:
        """Submits the task for review.

        **Parameters**

        - `assignee` (Optional[str]): User email to be assigned to the task whilst submitting the task.
        - `retain_assignee` (bool): Retains the current assignee of the task. This is ignored if `assignee` is provided. An Error will occur if the task does not already have an assignee and `retain_assignee` is True.
        - `bundle` (Optional[Bundle]): Optional bundle to be included with the submission.
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(
            stage_uuid,
            _ActionSubmit(task_uuid=self.uuid, assignee=assignee, retain_assignee=retain_assignee),
            bundle=bundle,
        )

    def assign(self, assignee: str, *, bundle: Optional[Bundle] = None) -> None:
        """Assigns the task to a user.

        **Parameters**

        - `assignee` (str): The email of the user to assign the task to.
        - `bundle` (Optional[Bundle]): Optional bundle to be included with the assignment.
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(stage_uuid, _ActionAssign(task_uuid=self.uuid, assignee=assignee), bundle=bundle)

    def release(self, *, bundle: Optional[Bundle] = None) -> None:
        """Releases the task from the current user.

        **Parameters**

        - `bundle` (Optional[Bundle]): Optional bundle to be included with the release.
        """
        workflow_client, stage_uuid = self._get_client_data()
        workflow_client.action(stage_uuid, _ActionRelease(task_uuid=self.uuid), bundle=bundle)



"""---
title: "Objects - Bundled Operations"
slug: "sdk-ref-objects-bundled-op"
hidden: false
metadata:
  title: "Objects - Bundled Operations"
  description: "Encord SDK Objects - Bundled Operations."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple

"""
Operation payloads to work with LabelRowV2 bundling.
These are internal helpers and not supposed to be used by external users.
"""


@dataclass
class BundledGetRowsPayload:
    uids: List[str]
    get_signed_url: bool
    include_object_feature_hashes: Optional[Set[str]]
    include_classification_feature_hashes: Optional[Set[str]]
    include_reviews: bool

    def add(self, other: BundledGetRowsPayload) -> BundledGetRowsPayload:
        self.uids.extend(other.uids)
        return self


@dataclass
class BundledCreateRowsPayload:
    uids: List[str]
    get_signed_url: bool
    branch_name: Optional[str]

    def add(self, other: BundledCreateRowsPayload) -> BundledCreateRowsPayload:
        self.uids.extend(other.uids)
        return self


@dataclass
class BundledSaveRowsPayload:
    uids: List[str]
    payload: List[Dict]
    validate_before_saving: Optional[bool]

    def add(self, other: BundledSaveRowsPayload) -> BundledSaveRowsPayload:
        self.uids.extend(other.uids)
        self.payload.extend(other.payload)
        self.validate_before_saving = self.validate_before_saving or other.validate_before_saving
        return self


@dataclass
class BundledSetPriorityPayload:
    priorities: List[Tuple[str, float]]

    def add(self, other: BundledSetPriorityPayload) -> BundledSetPriorityPayload:
        self.priorities.extend(other.priorities)
        return self


@dataclass
class BundledWorkflowCompletePayload:
    label_hashes: List[str]

    def add(self, other: BundledWorkflowCompletePayload) -> BundledWorkflowCompletePayload:
        self.label_hashes.extend(other.label_hashes)
        return self


@dataclass
class BundledWorkflowReopenPayload:
    label_hashes: List[str]

    def add(self, other: BundledWorkflowReopenPayload) -> BundledWorkflowReopenPayload:
        self.label_hashes.extend(other.label_hashes)
        return self



from datetime import datetime
from typing import Any, Dict
from uuid import UUID

from encord.common.enum import StringEnum
from encord.orm.base_dto import BaseDTO


class ClientMetadataSchemaTypes(StringEnum):
    NUMBER = "number"
    STRING = "string"
    BOOLEAN = "boolean"
    DATETIME = "datetime"
    GEOSPATIAL = "geospatial"
    ENUM = "enum"
    EMBEDDING = "embedding"
    LONG_STRING = "long_string"


class ClientMetadataSchema(BaseDTO):
    uuid: UUID
    metadata_schema: Dict[str, ClientMetadataSchemaTypes]
    organisation_id: int
    created_at: datetime
    updated_at: datetime


class ClientMetadataSchemaPayload(BaseDTO):
    metadata_schema: Dict[str, ClientMetadataSchemaTypes]

    def to_dict(self, by_alias=True, exclude_none=True) -> Dict[str, Any]:
        return self.metadata_schema



import json
from enum import Enum
from typing import Dict, Literal, Sequence, Union

from pydantic import BaseModel, Field
from typing_extensions import Annotated

from encord.http.v2.api_client import ApiClient

__all__ = ["MetadataSchema", "MetadataSchemaError", "MetadataSchemaScalarType"]

from encord.orm.base_dto import RootModelDTO


class _NumberFormat(Enum):
    F32 = "F32"
    F64 = "F64"


class _ClientMetadataSchemaTypeNumber(BaseModel):
    ty: Literal["number"] = "number"
    format: _NumberFormat = _NumberFormat.F64


class _ClientMetadataSchemaTypeBoolean(BaseModel):
    ty: Literal["boolean"] = "boolean"


class _ClientMetadataSchemaTypeVarChar(BaseModel):
    ty: Literal["varchar"] = "varchar"
    max_length: int = Field(8192, ge=0, le=8192)


class _ClientMetadataSchemaTypeDateTime(BaseModel):
    ty: Literal["datetime"] = "datetime"
    timezone: Literal["UTC"] = "UTC"


class _ClientMetadataSchemaTypeEnum(BaseModel):
    ty: Literal["enum"] = "enum"
    values: Sequence[str] = Field([], min_length=1, max_length=256)


class _ClientMetadataSchemaTypeEmbedding(BaseModel):
    ty: Literal["embedding"] = "embedding"
    size: int = Field(gt=0, le=4096)


class _ClientMetadataSchemaTypeText(BaseModel):
    ty: Literal["text"] = "text"


class _ClientMetadataSchemaTypeUUID(BaseModel):
    ty: Literal["uuid"] = "uuid"


class _ClientMetadataSchemaTypeVariantHint(Enum):
    NUMBER = "number"
    VARCHAR = "varchar"
    TEXT = "text"
    BOOLEAN = "boolean"
    DATETIME = "datetime"
    UUID = "uuid"

    def to_simple_str(self) -> Literal["boolean", "datetime", "uuid", "number", "varchar", "text", "uuid"]:
        if self.value == "number":
            return "number"
        elif self.value == "varchar":
            return "varchar"
        elif self.value == "text":
            return "text"
        elif self.value == "boolean":
            return "boolean"
        elif self.value == "datetime":
            return "datetime"
        elif self.value == "uuid":
            return "uuid"
        else:
            raise ValueError(f"Unknown simple schema type: {self}")

    @classmethod
    def _missing_(cls, value):
        if value in ("text", "long_string"):
            return cls.TEXT
        elif value in ("varchar", "string"):
            return cls.VARCHAR

        raise ValueError(f"Unknown simple schema type: {value}")


MetadataSchemaScalarType = _ClientMetadataSchemaTypeVariantHint


class _ClientMetadataSchemaTypeVariant(BaseModel):
    ty: Literal["variant"] = "variant"
    hint: _ClientMetadataSchemaTypeVariantHint = _ClientMetadataSchemaTypeVariantHint.TEXT


class _ClientMetadataSchemaTypeUser(BaseModel):
    ty: Literal["user"] = "user"


class _ClientMetadataSchemaTypeTombstoneDeletedTy(
    RootModelDTO[
        Annotated[
            Union[
                _ClientMetadataSchemaTypeNumber,
                _ClientMetadataSchemaTypeBoolean,
                _ClientMetadataSchemaTypeVarChar,
                _ClientMetadataSchemaTypeDateTime,
                _ClientMetadataSchemaTypeEnum,
                _ClientMetadataSchemaTypeEmbedding,
                _ClientMetadataSchemaTypeText,
                _ClientMetadataSchemaTypeUUID,
                _ClientMetadataSchemaTypeVariant,
                _ClientMetadataSchemaTypeUser,
            ],
            Field(discriminator="ty"),
        ]
    ]
):
    pass


class _ClientMetadataSchemaTypeTombstone(BaseModel):
    ty: Literal["tombstone"] = "tombstone"
    deleted_ty: Union[_ClientMetadataSchemaTypeTombstoneDeletedTy, None] = None


class _ClientMetadataSchemaOption(
    RootModelDTO[
        Annotated[
            Union[
                _ClientMetadataSchemaTypeNumber,
                _ClientMetadataSchemaTypeBoolean,
                _ClientMetadataSchemaTypeVarChar,
                _ClientMetadataSchemaTypeDateTime,
                _ClientMetadataSchemaTypeEnum,
                _ClientMetadataSchemaTypeEmbedding,
                _ClientMetadataSchemaTypeText,
                _ClientMetadataSchemaTypeUUID,
                _ClientMetadataSchemaTypeVariant,
                _ClientMetadataSchemaTypeUser,
                _ClientMetadataSchemaTypeTombstone,
            ],
            Field(discriminator="ty"),
        ]
    ]
):
    pass


class _ClientMetadataSchema(RootModelDTO[Dict[str, _ClientMetadataSchemaOption]]):
    """Internal type for a metadata schema."""

    pass


class MetadataSchemaError(RuntimeError):
    """Raised when an invariant is violated when mutating metadata schemas"""

    pass


def _is_valid_metadata_key(value: str) -> bool:
    if value.startswith("$") or value == "":
        return False
    for extra_char in ["-", "_", " ", "."]:
        value = value.replace(extra_char, "")
    if value == "":
        return False
    # deprecated special-keys
    if value in ("keyframes",):
        return False
    return value.isalnum()


def _assert_valid_metadata_key(value: str) -> None:
    if not _is_valid_metadata_key(value):
        raise MetadataSchemaError(f"{value} is an invalid format for a metadata key")


class MetadataSchema:
    """A class to manage the metadata schema for an organization.

    **Methods:**

    save() -> None: Saves the metadata schema to the backend if it has been modified.
    """

    _dirty: bool
    _schema: Dict[str, _ClientMetadataSchemaOption]
    _api_client: ApiClient

    def __init__(self, api_client: ApiClient) -> None:
        self._api_client = api_client
        schema_opt: _ClientMetadataSchema = api_client.get(  # type: ignore[type-var]
            "organisation/client-metadata-schema-v2",
            params=None,
            result_type=_ClientMetadataSchema,  # type: ignore[arg-type]
            allow_none=True,
        )
        self._schema = schema_opt.root if schema_opt is not None else {}  # type: ignore[assignment]
        self._dirty = False

    def save(self) -> None:
        """Saves the metadata schema to the backend if changes have been made."""
        if self._dirty:
            self._api_client.post(
                "organisation/client-metadata-schema-v2",
                params=None,
                payload=_ClientMetadataSchema.model_validate(self._schema),  # type: ignore[arg-type]
                result_type=None,
            )
            self._dirty = False

    def add_embedding(self, k: str, *, size: int) -> None:
        """Adds a new embedding to the metadata schema.

        **Parameters:**

        - k : str: The key under which the embedding will be stored in the schema.
        - size : int: The size of the embedding.

        **Raises:**

        MetadataSchemaError: If the key `k` is already defined in the schema.
        """
        if k in self._schema:
            raise MetadataSchemaError(f"{k} is already defined")
        _assert_valid_metadata_key(k)
        self._schema[k] = _ClientMetadataSchemaOption(root=_ClientMetadataSchemaTypeEmbedding(size=size))
        self._dirty = True

    def add_enum(self, k: str, *, values: Sequence[str]) -> None:
        """Adds a new enum to the metadata schema.

        **Parameters:**

        - k : str: The key under which the embedding will be stored in the schema.
        - values : Sequence[str]: The set of values for the enum (min 1, max 256).

        **Raises:**

        MetadataSchemaError: If the key `k` is already defined in the schema.
        """
        if k in self._schema:
            raise MetadataSchemaError(f"{k} is already defined")
        _assert_valid_metadata_key(k)
        if len(values) == 0:
            raise MetadataSchemaError("Must provide at least 1 value")
        elif len(values) >= 256:
            raise MetadataSchemaError("Cannot provide more than 256 values")
        self._schema[k] = _ClientMetadataSchemaOption(root=_ClientMetadataSchemaTypeEnum(values=sorted(values)))
        self._dirty = True

    def add_enum_options(self, k: str, *, values: Sequence[str]) -> None:
        """Adds extra valid enum values to an existing enum schema.

        **Parameters:**

        - k : str: The key referencing the enum.
        - values : Sequence[str]: The set of new values to add to the enum (min 1, max 256).

        **Raises:**

        MetadataSchemaError: If the key `k` is not defined in the schema or is not an enum.
        """
        if k not in self._schema:
            raise MetadataSchemaError(f"{k} is already defined")
        _assert_valid_metadata_key(k)
        v = self._schema[k].root
        if not isinstance(v, _ClientMetadataSchemaTypeEnum):
            raise MetadataSchemaError(f"{k} is not an enum")
        new_values = list(v.values) + list(values)
        if len(new_values) == 0 or len(values) == 0:
            raise MetadataSchemaError("Must provide at least 1 value")
        elif len(new_values) >= 256:
            raise MetadataSchemaError("Cannot provide more than 256 values")
        self._schema[k] = _ClientMetadataSchemaOption(root=_ClientMetadataSchemaTypeEnum(values=sorted(new_values)))
        self._dirty = True

    def add_scalar(
        self,
        k: str,
        *,
        data_type: Union[
            Literal["boolean", "datetime", "number", "uuid", "varchar", "text", "string", "long_string"],
            MetadataSchemaScalarType,
        ],
    ) -> None:
        """Sets a simple metadata type for a given key in the schema.

        **Parameters:**

        - k : str: The key for which the metadata type is being set.
        - data_type : Literal["boolean", "datetime", "number", "uuid", "varchar", "text", "string", "long_string"]
                   The type of metadata to be associated with the key. Must be a valid identifier.
                   "string" is an alias of "varchar"
                   "long_string" is an alias of "text"

        **Raises:**

        MetadataSchemaError: If the key `k` is already defined in the schema with a conflicting type.
        ValueError: If `data_type` is not a valid type of metadata identifier.
        """
        if isinstance(data_type, MetadataSchemaScalarType):
            data_type = data_type.to_simple_str()

        if k in self._schema:
            v = self._schema[k].root
            if not isinstance(v, (_ClientMetadataSchemaTypeVariant, _ClientMetadataSchemaTypeTombstone)):
                raise MetadataSchemaError(f"{k} is already defined")
            elif isinstance(v, _ClientMetadataSchemaTypeTombstone):
                if v.deleted_ty is None:
                    raise MetadataSchemaError(f"{k} is hard deleted")
                elif not isinstance(v.deleted_ty.root, _ClientMetadataSchemaTypeVariant):
                    raise MetadataSchemaError(f"{k} is cannot be restored with this type")
        _assert_valid_metadata_key(k)
        if data_type == "embedding":
            raise MetadataSchemaError("Embedding must be created explicitly")
        elif data_type == "enum":
            raise MetadataSchemaError("Enum must be created explicitly")
        self._schema[k] = _ClientMetadataSchemaOption(
            root=_ClientMetadataSchemaTypeVariant(hint=_ClientMetadataSchemaTypeVariantHint(data_type))
        )
        self._dirty = True

    def set_scalar(
        self,
        k: str,
        *,
        data_type: Union[
            Literal["boolean", "datetime", "number", "uuid", "varchar", "text", "string", "long_string"],
            MetadataSchemaScalarType,
        ],
    ) -> None:
        """Sets a simple metadata type for a given key in the schema.

        Alias of add_scalar

        **Parameters:**

        - k : str: The key for which the metadata type is being set.
        - data_type : Literal["boolean", "datetime", "number", "uuid", "varchar", "text", "string", "long_string"]
                   The type of metadata to be associated with the key. Must be a valid identifier.
                   "string" is an alias of "varchar"
                   "long_string" is an alias of "text"

        **Raises:**

        MetadataSchemaError: If the key `k` is already defined in the schema with a conflicting type.
        ValueError: If `data_type` is not a valid type of metadata identifier.
        """
        self.add_scalar(k, data_type=data_type)

    def delete_key(self, k: str, *, hard: bool = False) -> None:
        """Delete a metadata key from the schema.

        **Parameters:**

        k : str: The key for which the metadata type is being deleted.

        hard: bool: If the deletion should prevent indexing of this key unconditionally.
                Setting this to true prevents restoring the type definition in the future.

        **Raises:**

        MetadataSchemaError: If the key `k` is already deleted or not present in the schema
        """
        if k not in self._schema:
            raise MetadataSchemaError(f"{k} is not defined")
        _assert_valid_metadata_key(k)
        deleted_ty = self._schema[k].root
        if isinstance(deleted_ty, _ClientMetadataSchemaTypeTombstone):
            raise MetadataSchemaError(f"{k} is already deleted")
        self._schema[k] = _ClientMetadataSchemaOption(
            root=_ClientMetadataSchemaTypeTombstone(
                deleted_ty=None if hard else _ClientMetadataSchemaTypeTombstoneDeletedTy(root=deleted_ty)
            )
        )
        self._dirty = True

    def restore_key(self, k: str) -> None:
        """Restore a deleted metadata key to its original value.

        **Parameters:**

        k : str: The key for which the metadata type is to be restored.

        **Raises:**

        MetadataSchemaError: If the key `k` is not already deleted or not present in the schema
        """
        if k not in self._schema:
            raise MetadataSchemaError(f"{k} is not defined")
        v = self._schema[k]
        tombstone_ty = v.root
        if not isinstance(tombstone_ty, _ClientMetadataSchemaTypeTombstone):
            raise MetadataSchemaError(f"{k} is not currently deleted")
        if tombstone_ty.deleted_ty is None:
            raise MetadataSchemaError(f"{k} was hard deleted, this does not support restore")
        self._schema[k] = _ClientMetadataSchemaOption(root=tombstone_ty.deleted_ty.root)
        self._dirty = True

    def keys(self) -> Sequence[str]:
        """Returns a sequence of all keys defined in the metadata schema.

        **Returns:**

        Sequence[str]: A list of keys present in the metadata schema.
        """
        return list(self._schema.keys())

    def has_key(self, k: str) -> bool:
        """Check if any definition exists for a key.

        **Parameters:**

        k : str: The key for which the metadata type is to be retrieved.
        """
        return k in self._schema

    def is_key_deleted(self, k: str) -> bool:
        """Check if the key is defined as deleted. (Tombstone type)

        **Parameters:**

        k : str: The key for which the metadata type is to be retrieved.
        """
        if k not in self._schema:
            return False
        _assert_valid_metadata_key(k)
        v = self._schema[k].root
        return isinstance(v, _ClientMetadataSchemaTypeTombstone)

    def get_key_type(
        self, k: str
    ) -> Union[Literal["boolean", "datetime", "uuid", "number", "varchar", "text", "embedding", "enum"], None]:
        """Retrieves the metadata type associated with a given key.

        **Parameters:**

        k : str: The key for which the metadata type is to be retrieved.

        **Returns:**

        Literal["boolean", "datetime", "uuid", "number", "varchar", "text", "embedding", "enum"]: The metadata type associated with the key `k`.

        Raises:
        MetadataSchemaError: If the key `k` is not supported by the current SDK.
        """
        if k not in self._schema:
            return None
        _assert_valid_metadata_key(k)
        v = self._schema[k].root
        if isinstance(v, _ClientMetadataSchemaTypeEmbedding):
            return "embedding"
        elif isinstance(v, _ClientMetadataSchemaTypeEnum):
            return "enum"
        elif isinstance(v, _ClientMetadataSchemaTypeVariant):
            return v.hint.to_simple_str()
        elif isinstance(v, _ClientMetadataSchemaTypeTombstone):
            return None
        elif isinstance(v, _ClientMetadataSchemaTypeBoolean):
            return "boolean"
        elif isinstance(v, _ClientMetadataSchemaTypeNumber):
            return "number"
        elif isinstance(v, _ClientMetadataSchemaTypeDateTime):
            return "datetime"
        elif isinstance(v, _ClientMetadataSchemaTypeVarChar):
            return "varchar"
        elif isinstance(v, _ClientMetadataSchemaTypeText):
            return "text"
        elif isinstance(v, _ClientMetadataSchemaTypeUUID):
            return "uuid"
        else:
            raise MetadataSchemaError(f"{k} is not supported in the current SDK")

    def get_embedding_size(self, k: str) -> int:
        """Retrieves size associated with a given embedding.

        **Parameters:**

        k : str: The key for which the metadata type is to be retrieved.

        **Returns:**

        int: The size of the embedding

        **Raises:**

        MetadataSchemaError: If the key `k` is not defined in the schema or is not an embedding
        """
        if k not in self._schema:
            raise MetadataSchemaError(f"{k} is not defined")
        _assert_valid_metadata_key(k)
        v = self._schema[k].root
        if isinstance(v, _ClientMetadataSchemaTypeEmbedding):
            return v.size
        else:
            raise MetadataSchemaError(f"{k} does not refer to an embedding")

    def get_enum_options(self, k: str) -> Sequence[str]:
        """Retrieves all values associated with a given enum.

        **Parameters:**

        k : str: The key for which the metadata type is to be retrieved.

        **Returns:**

        Sequence[str]: The list of all values associated with an enum type.

        **Raises:**

        MetadataSchemaError: If the key `k` is not defined in the schema or is not an enum.
        """
        if k not in self._schema:
            raise MetadataSchemaError(f"{k} is not defined")
        _assert_valid_metadata_key(k)
        v = self._schema[k].root
        if isinstance(v, _ClientMetadataSchemaTypeEnum):
            return sorted(v.values)
        else:
            raise MetadataSchemaError(f"{k} does not refer to an enum")

    def __repr__(self):
        return json.dumps({k: self._schema[k].model_dump(mode="json") for k in sorted(self._schema.keys())})

    def __str__(self) -> str:
        schema = "Metadata Schema:\n"
        schema += "----------------\n"
        max_length = max([len(k) for k in self._schema.keys()], default=0)
        for k in sorted(self._schema.keys()):
            v = self._schema[k].root
            ty_hint_str: str
            padding: str = " " * (max_length - len(k))
            if isinstance(v, _ClientMetadataSchemaTypeEmbedding):
                ty_hint_str = f"embedding(size={v.size})"
            elif isinstance(v, _ClientMetadataSchemaTypeEnum):
                ty_hint_str = f"enum(values={sorted(v.values)})"
            elif isinstance(v, _ClientMetadataSchemaTypeVariant):
                ty_hint_str = f"scalar(hint={v.hint.to_simple_str()})"
            elif isinstance(v, _ClientMetadataSchemaTypeTombstone):
                continue
            elif isinstance(v, _ClientMetadataSchemaTypeBoolean):
                ty_hint_str = "boolean"
            elif isinstance(v, _ClientMetadataSchemaTypeNumber):
                ty_hint_str = "number"
            elif isinstance(v, _ClientMetadataSchemaTypeDateTime):
                ty_hint_str = "datetime"
            elif isinstance(v, _ClientMetadataSchemaTypeVarChar):
                ty_hint_str = "varchar"
            elif isinstance(v, _ClientMetadataSchemaTypeText):
                ty_hint_str = "text"
            elif isinstance(v, _ClientMetadataSchemaTypeUUID):
                ty_hint_str = "uuid"
            else:
                raise MetadataSchemaError(f"{k} is not supported in the current SDK")
            schema += f" - '{k}': {padding}{ty_hint_str}\n"

        return schema
